##############################################################################
#
# Copyright (c) 2008,2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
# pylint:disable=too-many-public-methods,unused-argument
from __future__ import absolute_import
from __future__ import print_function

import os
import tempfile

from ZODB.blob import remove_committed_dir
from ZODB.POSException import StorageTransactionError

from hamcrest import assert_that
from nti.testing.matchers import validly_provides

from relstorage.tests import TestCase

from ..interfaces import IBlobHelper

test_oid = b'\0' * 7 + b'\x01'
test_tid = b'\0' * 7 + b'\x02'


def write_file(fn, data):
    assert isinstance(data, str)
    with open(fn, 'w') as f:
        f.write(data)


def read_file(fn):
    with open(fn, 'r') as f:
        return f.read()


class BlobHelperTest(TestCase):

    shared_blob_dir = True
    move_method = 'vote'

    def setUp(self):
        self.uploaded = None
        self.blob_dir = tempfile.mkdtemp()

    def tearDown(self):
        remove_committed_dir(self.blob_dir)

    def _class(self):
        from relstorage.blobhelper import BlobHelper
        return BlobHelper

    def _make(self, *args, **kw):
        return self._class()(*args, **kw)

    def _make_default(self, cache_size=None,
                      download_action='write', keep_history=True):
        test = self

        class DummyOptions(object):
            blob_dir = self.blob_dir
            shared_blob_dir = self.shared_blob_dir
            blob_cache_size = cache_size

        class DummyMover(object):
            def download_blob(self, cursor, oid_int, tid_int, filename):
                if download_action == 'write':
                    write_file(filename, 'blob here')
                    return 9
                return 0

            def upload_blob(self, cursor, oid_int, tid_int, filename):
                test.uploaded = (oid_int, tid_int, filename)

        class DummyAdapter(object):
            mover = DummyMover()

            def __init__(self):
                self.keep_history = keep_history

        blobhelper = self._make(DummyOptions(), DummyAdapter())
        blobhelper.begin()
        return blobhelper

    def test_provides(self):
        assert_that(self._make_default(), validly_provides(IBlobHelper))

    def test_cannot_begin_twice(self):
        blobhelper = self._make_default()
        with self.assertRaises(StorageTransactionError):
            blobhelper.begin()

    def test_begin_after_abort(self):
        blobhelper = self._make_default()
        blobhelper.abort()
        blobhelper.begin()

    def test_begin_after_finish(self):
        blobhelper = self._make_default()
        blobhelper.finish(b'tid')
        blobhelper.begin()

    def test_loadBlob_shared_exists(self):
        blobhelper = self._make_default()
        fn = blobhelper.fshelper.getBlobFilename(test_oid, test_tid)
        os.makedirs(os.path.dirname(fn))
        write_file(fn, 'blob here')
        res = blobhelper.loadBlob(None, test_oid, test_tid)
        self.assertEqual(fn, res)

    def test_storeBlob_replace(self):
        called = []
        dummy_txn = object()

        def store_func(oid, tid, data, txn):
            called.append((oid, tid, data, txn))

        fn1 = os.path.join(self.blob_dir, 'newblob')
        write_file(fn1, 'here a blob')
        fn2 = os.path.join(self.blob_dir, 'newblob2')
        write_file(fn2, 'there a blob')

        blobhelper = self._make_default()
        self.assertFalse(blobhelper.txn_has_blobs)
        blobhelper.storeBlob(None, store_func, test_oid, test_tid, 'blob pickle',
                             fn1, '', dummy_txn)
        blobhelper.storeBlob(None, store_func, test_oid, test_tid, 'blob pickle',
                             fn2, '', dummy_txn)
        self.assertFalse(os.path.exists(fn1))
        self.assertFalse(os.path.exists(fn2))
        self.assertTrue(blobhelper.txn_has_blobs)
        target_fn = blobhelper._txn_blobs[test_oid]
        self.assertEqual(read_file(target_fn), 'there a blob')

    def test_move_into_place(self):
        blobhelper = self._make_default()
        d = blobhelper.fshelper.getPathForOID(test_oid, create=True)
        fn1 = os.path.join(d, 'newblob')
        write_file(fn1, 'here a blob')
        blobhelper._txn_blobs = {test_oid: fn1}
        move_method = getattr(blobhelper, self.move_method)
        move_method(test_tid)
        fn2 = blobhelper.fshelper.getBlobFilename(test_oid, test_tid)
        self.assertEqual(read_file(fn2), 'here a blob')

    def test_abort(self):
        blobhelper = self._make_default()
        d = blobhelper.fshelper.getPathForOID(test_oid, create=True)
        fn1 = os.path.join(d, 'newblob')
        write_file(fn1, 'here a blob')
        blobhelper._txn_blobs = {test_oid: fn1}
        blobhelper.abort()
        fn2 = blobhelper.fshelper.getBlobFilename(test_oid, test_tid)
        self.assertFalse(os.path.exists(fn1))
        self.assertFalse(os.path.exists(fn2))


class NoBlobHelperTest(TestCase):

    def _class(self):
        from relstorage.blobhelper import NoBlobHelper
        return NoBlobHelper

    def makeOne(self):
        return self._class()()

    def test_provides(self):
        assert_that(self.makeOne(), validly_provides(IBlobHelper))



from __future__ import absolute_import
from __future__ import print_function

import unittest

from relstorage.zodburi_resolver import RelStorageURIResolver

from . import mock

class AbstractURIResolverTestBase(unittest.TestCase):

    adapter_name = None
    prefix = ''

    def _get_helper(self):
        raise NotImplementedError()

    def _makeOne(self):
        helper = self._get_helper()
        return RelStorageURIResolver(helper())

    def setUp(self):
        # relstorage.options.Options is little more than a dict.
        # We make it comparable to simplify the tests.
        from relstorage.options import Options
        Options.__eq__ = lambda s, o: vars(s) == vars(o)

        self.patcher1 = mock.patch('relstorage.zodburi_resolver.RelStorage')
        self.patcher2 = mock.patch(self.adapter_name)
        self.RelStorage = self.patcher1.start()
        try:
            self.DBAdapter = self.patcher2.start()
        except ImportError as e:
            raise unittest.SkipTest(str(e))

    def tearDown(self):
        self.patcher1.stop()
        self.patcher2.stop()

    def test_bool_args(self):
        resolver = self._makeOne()
        f = resolver.interpret_kwargs
        kwargs = f({'read_only':'1'})
        self.assertEqual(kwargs[0], {'read_only':1})
        kwargs = f({'read_only':'true'})
        self.assertEqual(kwargs[0], {'read_only':1})
        kwargs = f({'read_only':'on'})
        self.assertEqual(kwargs[0], {'read_only':1})
        kwargs = f({'read_only':'off'})
        self.assertEqual(kwargs[0], {'read_only':0})
        kwargs = f({'read_only':'no'})
        self.assertEqual(kwargs[0], {'read_only':0})
        kwargs = f({'read_only':'false'})
        self.assertEqual(kwargs[0], {'read_only':0})

    def _format_db(self, dbname='somedb', user='someuser', password='somepass',
                   host='somehost', port='5432', **kwargs):
        raise NotImplementedError()

    def test_call(self):
        from relstorage.options import Options
        resolver = self._makeOne()
        factory, _dbkw = resolver(
            self.prefix + '://someuser:somepass@somehost:5432/somedb'
            '?read_only=1&cache_servers=123,456')
        factory()

        expected_options = Options(read_only=1, cache_servers=('123', '456'))
        self.DBAdapter.assert_called_once_with(
            options=expected_options, **self._format_db())
        self.RelStorage.assert_called_once_with(
            adapter=self.DBAdapter(), options=expected_options)

    def test_call_adapter_options(self):
        from relstorage.options import Options
        resolver = self._makeOne()
        factory, _dbkw = resolver(
            self.prefix + '://someuser:somepass@somehost:5432/somedb'
            '?read_only=1&connect_timeout=10')
        factory()

        expected_options = Options(read_only=1)
        self.DBAdapter.assert_called_once_with(
            options=expected_options,
            **self._format_db(connect_timeout=10))
        self.RelStorage.assert_called_once_with(
            adapter=self.DBAdapter(), options=expected_options)

    def test_invoke_factory_driver_auto(self, driver_name='auto'):
        from relstorage.options import Options
        resolver = self._makeOne()
        factory, _dbkw = resolver(
            self.prefix + '://someuser:somepass@somehost:5432/somedb'
            '?driver=' + driver_name
        )
        factory()
        expected_options = Options(driver=driver_name)
        self.DBAdapter.assert_called_once_with(
            options=expected_options,
            **self._format_db())

    def test_invoke_factory_demostorage(self):
        from ZODB.DemoStorage import DemoStorage
        resolver = self._makeOne()
        factory, _dbkw = resolver(
            self.prefix + '://someuser:somepass@somehost:5432/somedb'
            '?read_only=1&demostorage=true')
        self.assertTrue(isinstance(factory(), DemoStorage))

    def test_dbargs(self):
        resolver = self._makeOne()
        _factory, dbkw = resolver(
            self.prefix + '://someuser:somepass@somehost:5432/somedb'
            '?read_only=1&connection_pool_size=1&connection_cache_size=1'
            '&database_name=dbname')
        self.assertEqual(dbkw, {'connection_pool_size': '1',
                                'connection_cache_size': '1',
                                'database_name': 'dbname'})


class TestPostgreSQLURIResolver(AbstractURIResolverTestBase):
    adapter_name = 'relstorage.adapters.postgresql.PostgreSQLAdapter'
    prefix = 'postgres'

    def _get_helper(self):
        from relstorage.zodburi_resolver import PostgreSQLAdapterHelper
        return PostgreSQLAdapterHelper


    def _format_db(self, dbname='somedb', user='someuser', password='somepass',
                   host='somehost', port='5432', **kwargs):
        dsn = ("dbname='%s' user='%s' password='%s' host='%s' port='%s'"
               % (dbname, user, password, host, port))
        if 'connect_timeout' in kwargs:
            dsn += " connect_timeout='%s'" % kwargs['connect_timeout']
        else:
            assert not kwargs

        return {'dsn': dsn}


class TestMySQLURIResolver(AbstractURIResolverTestBase):
    adapter_name = 'relstorage.adapters.mysql.MySQLAdapter'
    prefix = 'postgres'

    def _get_helper(self):
        from relstorage.zodburi_resolver import MySQLAdapterHelper
        return MySQLAdapterHelper

    def _format_db(self, dbname='somedb', user='someuser', password='somepass',
                   host='somehost', port='5432', **kwargs):
        args = dict(locals())
        args.update(kwargs)
        args['db'] = args['dbname']; del args['dbname']
        args['passwd'] = args['password']; del args['password']
        args['port'] = int(args['port'])

        del args['kwargs']
        del args['self']
        return args

class TestOracleURIResolver(AbstractURIResolverTestBase):
    adapter_name = 'relstorage.adapters.oracle.OracleAdapter'
    prefix = 'oracle'

    def _get_helper(self):
        from relstorage.zodburi_resolver import OracleAdapterHelper
        return OracleAdapterHelper

    # The default URI the test uses puts user and password and host into
    # the URI proper, but oracle expects them in the query params.
    def _format_db(self, dbname='somedb', user='someuser', password='somepass',
                   host='somehost', port='5432', **kwargs):
        return {}

class TestSQLiteURIResolver(AbstractURIResolverTestBase):
    adapter_name = 'relstorage.adapters.sqlite.adapter.Sqlite3Adapter'
    prefix = 'sqlite'

    def _get_helper(self):
        from relstorage.zodburi_resolver import SqliteAdapterHelper
        return SqliteAdapterHelper

    def _format_db(self, dbname='somedb', user='someuser', password='somepass',
                   host='somehost', port='5432', **kwargs):
        return {}


del AbstractURIResolverTestBase # So it doesn't get discovered as a test

class TestEntryPoints(unittest.TestCase):

    def _check_entry_point(self, name, cls, helper_cls):
        from pkg_resources import load_entry_point

        target = load_entry_point('relstorage', 'zodburi.resolvers', name)
        self.assertIsInstance(target, cls)
        self.assertIsInstance(target.adapter_helper, helper_cls)

    def test_postgres(self):
        from relstorage.zodburi_resolver import PostgreSQLAdapterHelper
        self._check_entry_point('postgres', RelStorageURIResolver, PostgreSQLAdapterHelper)

    def test_mysql(self):
        from relstorage.zodburi_resolver import MySQLAdapterHelper
        self._check_entry_point('mysql', RelStorageURIResolver, MySQLAdapterHelper)

    def test_oracle(self):
        from relstorage.zodburi_resolver import OracleAdapterHelper
        self._check_entry_point('oracle', RelStorageURIResolver, OracleAdapterHelper)

    def test_sqlite(self):
        from relstorage.zodburi_resolver import SqliteAdapterHelper
        self._check_entry_point('sqlite', RelStorageURIResolver, SqliteAdapterHelper)



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
Test mixin dealing with different locking scenarios.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import threading
import time

from functools import partial
from functools import update_wrapper

import transaction

from ZODB.DB import DB
from ZODB.Connection import TransactionMetaData

from ZODB.tests.MinPO import MinPO

from relstorage.storage.interfaces import VoteReadConflictError

from . import TestCase
from . import skipIfNoConcurrentWriters

def WithAndWithoutInterleaving(func):
    # Expands a test case into two tests, for those that can run
    # both with the stored procs and without it.
    def _interleaved(self):
        adapter = self._storage._adapter
        if not adapter.DEFAULT_LOCK_OBJECTS_AND_DETECT_CONFLICTS_INTERLEAVABLE:
            adapter.force_lock_objects_and_detect_conflicts_interleavable = True

        func(self)
        if 'force_lock_objects_and_detect_conflicts_interleavable' in adapter.__dict__:
            del adapter.force_lock_objects_and_detect_conflicts_interleavable

    def _stored_proc(self):
        adapter = self._storage._adapter
        if adapter.DEFAULT_LOCK_OBJECTS_AND_DETECT_CONFLICTS_INTERLEAVABLE:
            # No stored proc version.
            return
        func(self)

    def test(self):
        # Sadly using self.subTest()
        # causes zope-testrunner to lose the exception reports.
        _stored_proc(self)
        _interleaved(self)

    return update_wrapper(test, func)


class TestLocking(TestCase):
    # pylint:disable=abstract-method

    _storage = None

    def make_storage(self, *args, **kwargs):
        raise NotImplementedError

    @property
    def __tiny_commit_time(self):
        # Use a very small commit lock timeout.
        if self._storage._adapter.schema.database_type == 'postgresql':
            return 0.1
        return 1

    def __store_two_for_read_current_error(self, release_extra_storage=False):
        db = self._closing(DB(self._storage, pool_size=1))
        conn = db.open()
        root = conn.root()
        root['object1'] = MinPO('object1')
        root['object2'] = MinPO('object2')
        transaction.commit()

        obj1_oid = root['object1']._p_oid
        obj2_oid = root['object2']._p_oid
        obj1_tid = root['object1']._p_serial
        assert obj1_tid == root['object2']._p_serial

        conn.close()
        # We can't close the DB, that will close the storage that we
        # still need. But we can release its storage, since we'll never use
        # this again.
        if release_extra_storage:
            conn._normal_storage.release()
        return obj1_oid, obj2_oid, obj1_tid, db

    def __read_current_and_lock(self, storage, read_current_oid, lock_oid, tid,
                                begin=True, tx=None):
        tx = tx if tx is not None else TransactionMetaData()
        if begin:
            storage.tpc_begin(tx)
        if read_current_oid is not None:
            storage.checkCurrentSerialInTransaction(read_current_oid, tid, tx)
        storage.store(lock_oid, tid, b'bad pickle2', '', tx)
        storage.tpc_vote(tx)
        return tx

    def __do_check_error_with_conflicting_concurrent_read_current(
            self,
            exception_in_b,
            commit_lock_timeout=None,
            storageA=None,
            storageB=None,
            identical_pattern_a_b=False,
            copy_interleave=('A', 'B'),
            abort=True
    ):
        # pylint:disable=too-many-locals
        root_adapter = self._storage._adapter
        if commit_lock_timeout:
            root_adapter.locker.commit_lock_timeout = commit_lock_timeout
            self._storage._options.commit_lock_timeout = commit_lock_timeout

        if storageA is None:
            storageA = self._closing(self._storage.new_instance())
        if storageB is None:
            storageB = self._closing(self._storage.new_instance())

        should_ileave = root_adapter.force_lock_objects_and_detect_conflicts_interleavable
        if 'A' in copy_interleave:
            storageA._adapter.force_lock_objects_and_detect_conflicts_interleavable = should_ileave
        if 'B' in copy_interleave:
            storageB._adapter.force_lock_objects_and_detect_conflicts_interleavable = should_ileave

        # First, store the two objects in an accessible location.
        obj1_oid, obj2_oid, tid, _db = self.__store_two_for_read_current_error(
            release_extra_storage=True)
        # XXX: We'd like to close the DB here, but the best we can do is to release the
        # extra storage it made.


        # Now transaction A readCurrent 1 and modify 2
        # up through the vote phase
        txa = self.__read_current_and_lock(storageA, obj1_oid, obj2_oid, tid)

        if not identical_pattern_a_b:
            # Second transaction does exactly the opposite, and blocks,
            # raising an exception (usually)
            lock_b = partial(
                self.__read_current_and_lock,
                storageB,
                obj2_oid,
                obj1_oid,
                tid
            )
        else:
            lock_b = partial(
                self.__read_current_and_lock,
                storageB,
                obj1_oid,
                obj2_oid,
                tid
            )

        txb = None
        before = time.time()
        if exception_in_b:
            with self.assertRaises(exception_in_b):
                lock_b()
        else:
            txb = lock_b()
        after = time.time()
        duration_blocking = after - before

        if abort:
            storageA.tpc_abort(txa)
            storageB.tpc_abort(txb)

        return duration_blocking

    @WithAndWithoutInterleaving
    def checkTL_ConflictingReadCurrent(self):
        # Given two objects 1 and 2, if transaction A does readCurrent(1)
        # and modifies 2 up through the voting phase, and then transaction
        # B does precisely the same thing,
        # we get an error after ``commit_lock_timeout``  and not a deadlock.
        from relstorage.adapters.interfaces import UnableToLockRowsToModifyError
        from relstorage.adapters.interfaces import UnableToLockRowsToReadCurrentError

        commit_lock_timeout = self.__tiny_commit_time
        duration_blocking = self.__do_check_error_with_conflicting_concurrent_read_current(
            (UnableToLockRowsToModifyError
             # Oracle has no share locks so we can't distinguish.
             if not self.__is_oracle()
             else UnableToLockRowsToReadCurrentError),
            commit_lock_timeout=commit_lock_timeout,
            identical_pattern_a_b=True
        )

        self.assertLessEqual(duration_blocking, commit_lock_timeout * 3)

    @WithAndWithoutInterleaving
    def checkTL_ReadCurrentConflict_DoesNotTakeExclusiveLocks(self):
        # Proves that if we try to check an old serial that has already moved on,
        # we don't try taking exclusive locks at all.
        from relstorage.adapters.interfaces import UnableToLockRowsToModifyError
        obj1_oid, obj2_oid, tid, db = self.__store_two_for_read_current_error()
        assert obj1_oid, obj1_oid
        assert obj2_oid, obj2_oid
        assert tid, tid

        root_adapter = self._storage._adapter
        commit_lock_timeout = self.__tiny_commit_time
        root_adapter.locker.commit_lock_timeout = commit_lock_timeout
        self._storage._options.commit_lock_timeout = commit_lock_timeout

        storageA = self._closing(self._storage.new_instance())
        storageC = self._closing(self._storage.new_instance())

        # The Begin phase actually calls into the database when readCurrent()
        # is called to verify the tid. So we actually need to do that much of it now.
        storageB = self._closing(self._storage.new_instance())
        txb = TransactionMetaData()
        storageB.tpc_begin(txb) # Capture our current tid now
        storageB.checkCurrentSerialInTransaction(obj2_oid, tid, txb)

        should_ileave = root_adapter.force_lock_objects_and_detect_conflicts_interleavable
        for s in storageA, storageB, storageC:
            s._adapter.force_lock_objects_and_detect_conflicts_interleavable = should_ileave

        # Walk through a full transaction in A so that the tid changes.
        conn = db.open()
        root = conn.root()
        root['object1'].some_attr = 1
        root['object2'].some_attr = 1
        transaction.commit()
        new_tid = root['object1']._p_serial
        conn.close()
        self.assertGreater(new_tid, tid)

        # Tx A takes exclusive lock on object1
        txa = self.__read_current_and_lock(storageA, None, obj1_oid, new_tid)

        # Prove that it's locked. (This is slow on MySQL 5.7)
        with self.assertRaises(UnableToLockRowsToModifyError):
            self.__read_current_and_lock(storageC, None, obj1_oid, new_tid)

        # Now, try to readCurrent of object2 in the old tid, and take an exclusive lock
        # on obj1. We should immediately get a read current error and not conflict with the
        # exclusive lock.
        with self.assertRaisesRegex(VoteReadConflictError, "serial this txn started"):
            self.__read_current_and_lock(storageB, obj2_oid, obj1_oid, tid, begin=False, tx=txb)

        # Which is still held because we cannot lock it.
        with self.assertRaises(UnableToLockRowsToModifyError):
            self.__read_current_and_lock(storageC, None, obj1_oid, new_tid)

        storageA.tpc_abort(txa)

    @skipIfNoConcurrentWriters
    @WithAndWithoutInterleaving
    def checkTL_OverlappedReadCurrent_SharedLocksFirst(self):
        # Starting with two objects 1 and 2, if transaction A modifies 1 and
        # does readCurrent of 2, up through the voting phase, and transaction B does
        # exactly the opposite, transaction B is immediately killed with a read conflict
        # error. (We use the same two objects instead of a new object in transaction B to prove
        # shared locks are taken first.)
        from relstorage.adapters.interfaces import UnableToLockRowsToReadCurrentError
        from relstorage.tests.util import RUNNING_ON_CI
        commit_lock_timeout = self.__tiny_commit_time
        duration_blocking = self.__do_check_error_with_conflicting_concurrent_read_current(
            UnableToLockRowsToReadCurrentError,
            commit_lock_timeout=commit_lock_timeout,
        )
        # The NOWAIT lock should be very quick to fire.
        assert self._storage._adapter.locker.supports_row_lock_nowait is not None
        if self._storage._adapter.locker.supports_row_lock_nowait:
            multiplier = 1.3
            if RUNNING_ON_CI and 'mysql' in type(self).__name__:
                # On Travis CI, we observe MySQL 8.0.21 to be very slow at this
                # for some reason.
                multiplier = 2.1
            self.assertLessEqual(duration_blocking, commit_lock_timeout * multiplier)
        else:
            # Sigh. Old MySQL. Very slow. This takes around 4.5s to run both iterations.
            self.assertLessEqual(duration_blocking, commit_lock_timeout * 2.5)


    def __lock_rows_being_modified_only(self, storage, cursor,
                                        _current_oids, _share_blocks,
                                        _after_share=lambda: None):
        # A monkey-patch for Locker.lock_current_objects to only take the exclusive
        # locks.
        storage._adapter.locker._lock_rows_being_modified(cursor)

    def __assert_small_blocking_duration(self, storage, duration_blocking):
        # Even though we just went with the default commit_lock_timeout,
        # which is large...
        self.assertGreaterEqual(storage._options.commit_lock_timeout, 10)
        # ...the lock violation happened very quickly...
        # except on Oracle, which takes closer to the actual timeout , for some reason...
        # possibly because we don't have share locks?
        duration = 3
        if self.__is_oracle():
            duration = 12

        self.assertLessEqual(duration_blocking, duration)

    @skipIfNoConcurrentWriters
    def checkTL_InterleavedConflictingReadCurrent(self):
        # Similar to
        # ``checkTL_ConflictingReadCurrent``
        # except that we pause the process immediately after txA takes
        # its exclusive locks to let txB take *its* exclusive locks.
        # Then txB can go for the shared locks, which will block if
        # we're not in ``NOWAIT`` mode for shared locks. This tests
        # that we don't block, we report a retryable error. Note that
        # we don't adjust the commit_lock_timeout; it doesn't apply.
        #
        # If we're using a stored procedure, this test will break
        # because we won't be able to force the interleaving, so we make it
        # use the old version.

        from relstorage.adapters.interfaces import UnableToLockRowsToReadCurrentError
        storageA = self._closing(self._storage.new_instance())
        # This turns off stored procs and lets us control that phase.
        storageA._adapter.force_lock_objects_and_detect_conflicts_interleavable = True

        storageA._adapter.locker.lock_current_objects = partial(
            self.__lock_rows_being_modified_only,
            storageA)

        duration_blocking = self.__do_check_error_with_conflicting_concurrent_read_current(
            UnableToLockRowsToReadCurrentError,
            storageA=storageA,
            copy_interleave=()
        )

        self.__assert_small_blocking_duration(storageA, duration_blocking)

    @skipIfNoConcurrentWriters
    def checkTL_InterleavedConflictingReadCurrentDeadlock(self):
        # pylint:disable=too-many-statements
        # Like
        # ``checkTL_InterleavedConflictingReadCurrent``
        # except that we interleave both txA and txB: txA takes modify
        # lock, txB takes modify lock, txA attempts shared lock, txB
        # attempts shared lock. This results in a database deadlock, which is reported as
        # a retryable error.
        #
        # We have to use a thread to do the shared locks because it blocks.
        from relstorage.adapters.interfaces import UnableToLockRowsToReadCurrentError
        from relstorage.storage.tpc.vote import AbstractVote as VotePhase
        from relstorage.storage.tpc import NotInTransaction

        store_conn_pool = self._storage._store_connection_pool
        self.assertEqual(store_conn_pool.instance_count, 1)

        storageA = self._closing(self._storage.new_instance())
        storageB = self._closing(self._storage.new_instance())
        storageA.last_error = storageB.last_error = None

        self.assertEqual(store_conn_pool.instance_count, 3)

        storageA._adapter.locker.lock_current_objects = partial(
            self.__lock_rows_being_modified_only,
            storageA)
        storageB._adapter.locker.lock_current_objects = partial(
            self.__lock_rows_being_modified_only,
            storageB)

        # This turns off stored procs for locking.
        storageA._adapter.force_lock_readCurrent_for_share_blocking = True
        storageB._adapter.force_lock_readCurrent_for_share_blocking = True

        # This won't actually block, we haven't conflicted yet.
        self.__do_check_error_with_conflicting_concurrent_read_current(
            None,
            storageA=storageA,
            storageB=storageB,
            abort=False
        )
        self.assertEqual(store_conn_pool.instance_count, 3)

        self.assertIsInstance(storageA._tpc_phase, VotePhase)
        self.assertIsInstance(storageB._tpc_phase, VotePhase)

        cond = threading.Condition()
        cond.acquire()
        def lock_shared(storage, notify=True):
            cursor = storage._tpc_phase.shared_state.store_connection.cursor
            read_current_oids = storage._tpc_phase.required_tids.keys()
            if notify:
                cond.acquire(5)
                cond.notifyAll()
                cond.release()

            try:
                storage._adapter.locker._lock_readCurrent_oids_for_share(
                    cursor,
                    read_current_oids,
                    shared_locks_block=True
                )
            except UnableToLockRowsToReadCurrentError as ex:
                storage.last_error = ex
            finally:
                if notify:
                    cond.acquire(5)
                    cond.notifyAll()
                    cond.release()


        thread_locking_a = threading.Thread(
            target=lock_shared,
            args=(storageA,),
            name="thread_locking_a",
        )
        thread_locking_a.start()

        # wait for the background thread to get ready to lock.
        cond.acquire(5)
        cond.wait(5)
        cond.release()

        begin = time.time()

        lock_shared(storageB, notify=False)

        # Wait for background thread to finish.
        cond.acquire(5)
        cond.wait(5)
        cond.release()

        end = time.time()
        duration_blocking = end - begin

        # Now, one or the other storage got killed by the deadlock
        # detector, but NOT both. Which one depends on the database.
        # PostgreSQL likes to kill the foreground thread (storageB),
        # MySQL likes to kill the background thread (storageA)...
        # And Oracle likes to kill them both! Argh!
        self.assertTrue(storageA.last_error or storageB.last_error,
                        (storageA.last_error, storageB.last_error))
        if not self.__is_oracle():
            self.assertFalse(storageA.last_error and storageB.last_error,
                             (storageA.last_error, storageB.last_error))

        last_error = storageA.last_error or storageB.last_error

        self.assertIn('deadlock', str(last_error).lower())

        self.__assert_small_blocking_duration(storageA, duration_blocking)
        self.__assert_small_blocking_duration(storageB, duration_blocking)

        storageA.tpc_abort(storageA._tpc_phase.transaction)
        storageB.tpc_abort(storageB._tpc_phase.transaction)

        self.assertIsInstance(storageA._tpc_phase, NotInTransaction)
        self.assertIsInstance(storageB._tpc_phase, NotInTransaction)

        self.assertEqual(store_conn_pool.instance_count, 3)
        storageA.release()
        storageB.release()

        self.assertLessEqual(store_conn_pool.instance_count, 1)
        self.assertLessEqual(store_conn_pool.pooled_connection_count, 1)

    def __is_oracle(self):
        # This is an anti-pattern. we really should subclass the tests
        # in our testoracle.py file.
        return self._storage._adapter.schema.database_type == 'oracle'



"""Tests of relstorage.blobhelper"""
# pylint:disable=too-many-public-methods,unused-argument

import os

from . import test_blobhelper
from .test_blobhelper import write_file
from .test_blobhelper import read_file
from .test_blobhelper import test_oid
from .test_blobhelper import test_tid


class SharedBlobHelperTest(test_blobhelper.BlobHelperTest):
    # Tests that only apply to shared blob dirs

    shared_blob_dir = True

    def test_ctor_with_shared_blob_dir(self):
        blobhelper = self._make_default()
        self.assertIsNotNone(blobhelper.fshelper)
        self.assertFalse(hasattr(blobhelper, 'cache_checker'))
        from ZODB.blob import LAYOUTS
        self.assertEqual(blobhelper.fshelper.layout, LAYOUTS['bushy'])

    def test_openCommittedBlobFile_retry_fail_on_shared(self):
        loadBlob_calls = []

        blobhelper = self._make_default()
        orig_loadBlobInternal = blobhelper._loadBlobInternal
        def loadBlob_wrapper(cursor, oid, serial, open_lock):
            loadBlob_calls.append(1)
            fn = orig_loadBlobInternal(cursor, oid, serial, open_lock)
            os.remove(fn)
            return fn

        fn = blobhelper.fshelper.getBlobFilename(test_oid, test_tid)
        os.makedirs(os.path.dirname(fn))
        write_file(fn, 'blob here')

        blobhelper._loadBlobInternal = loadBlob_wrapper

        from ZODB.POSException import POSKeyError
        with self.assertRaises(POSKeyError):
            blobhelper.openCommittedBlobFile(None, test_oid, test_tid)
        self.assertEqual(loadBlob_calls, [1, 1])

    def test_loadBlob_shared_missing(self):
        blobhelper = self._make_default()
        from ZODB.POSException import POSKeyError
        with self.assertRaises(POSKeyError):
            blobhelper.loadBlob(None, test_oid, test_tid)

    def test_storeBlob_shared(self):
        called = []
        dummy_txn = object()

        def store_func(oid, tid, data, txn):
            called.append((oid, tid, data, txn))

        fn = os.path.join(self.blob_dir, 'newblob')
        write_file(fn, 'here a blob')

        blobhelper = self._make_default()
        self.assertFalse(blobhelper.txn_has_blobs)
        blobhelper.storeBlob(None, store_func, test_oid, test_tid, 'blob pickle',
                             fn, '', dummy_txn)
        self.assertFalse(os.path.exists(fn))
        self.assertTrue(blobhelper.txn_has_blobs)
        self.assertEqual(called,
                         [(test_oid, test_tid, 'blob pickle', dummy_txn)])

    def test_restoreBlob_shared(self):
        fn = os.path.join(self.blob_dir, 'newblob')
        write_file(fn, 'here a blob')
        blobhelper = self._make_default()
        blobhelper.restoreBlob(None, test_oid, test_tid, fn)
        self.assertFalse(os.path.exists(fn))
        target_fn = blobhelper.fshelper.getBlobFilename(test_oid, test_tid)
        self.assertEqual(read_file(target_fn), 'here a blob')

    def test_copy_undone_shared(self):
        blobhelper = self._make_default()
        copied = [(1, 1), (11, 1)]
        fn = blobhelper.fshelper.getBlobFilename(test_oid, test_oid)
        os.makedirs(os.path.dirname(fn))
        write_file(fn, 'blob here')
        blobhelper.copy_undone(copied, test_tid)
        self.assertTrue(blobhelper.txn_has_blobs)
        fn2 = blobhelper.fshelper.getBlobFilename(test_oid, test_tid)
        self.assertEqual(read_file(fn2), 'blob here')

    def test_after_pack_shared_with_history(self):
        blobhelper = self._make_default()
        fn = blobhelper.fshelper.getBlobFilename(test_oid, test_tid)
        os.makedirs(os.path.dirname(fn))
        write_file(fn, 'blob here')
        blobhelper.after_pack(1, 2)
        self.assertFalse(os.path.exists(fn))

    def test_after_pack_shared_without_history(self):
        blobhelper = self._make_default(keep_history=False)
        fn = blobhelper.fshelper.getBlobFilename(test_oid, test_tid)
        os.makedirs(os.path.dirname(fn))
        write_file(fn, 'blob here')
        blobhelper.after_pack(1, 2)
        self.assertFalse(os.path.exists(fn))



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2016 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
MySQL IDBDriver implementations.
"""
from __future__ import absolute_import
from __future__ import print_function

from ...drivers import AbstractModuleDriver
from ...drivers import implement_db_driver_options
from ...sql import Compiler
from ...sql import DefaultDialect
from ...sql import OID
from ...sql import TID
from ...sql import BinaryString
from ...sql import State

database_type = 'mysql'

class MySQLCompiler(Compiler):

    def can_prepare(self):
        # If there are params, we can't prepare unless we're using
        # the binary protocol; otherwise we have to SET user variables
        # with extra round trips, which is worse.
        return not self.placeholders and super(MySQLCompiler, self).can_prepare()

    _PREPARED_CONJUNCTION = 'FROM'

    def _prepared_param(self, number):
        return '?'

    def _quote_query_for_prepare(self, query):
        return "'{query}'".format(query=query)

    # Upserts: It's a good idea to use `ON DUPLICATE KEY UPDATE`
    # instead of REPLACE because `ON DUPLICATE` updates rows in place
    # instead of performing a DELETE followed by an INSERT; that might
    # matter for row locking or MVCC, depending on isolation level and
    # locking strategies, and it's been said that it matters for
    # backend IO, especially on things like a large distributed SAN
    # (https://github.com/zodb/relstorage/issues/189). This is also
    # more similar to what PostgreSQL uses, possibly allowing more
    # query sharing (with a smarter query runner/interpreter).


    def visit_upsert_conflict_column(self, _column):
        self.emit_keyword('ON DUPLICATE KEY')

    def visit_upsert_conflict_update(self, update):
        self.emit_keyword('UPDATE')
        self.visit_csv(update.col_expressions)

    def visit_upsert_excluded_column(self, column):
        self.emit('VALUES(%s)' % (column.name,))


class MySQLDialect(DefaultDialect):

    datatype_map = dict(DefaultDialect.datatype_map)
    datatype_map.update({
        OID: 'BIGINT UNSIGNED',
        TID: 'BIGINT UNSIGNED',
        BinaryString: 'BLOB',
        #MD5: 'CHAR(32) CHARACTER SET ascii',
        State: 'LONGBLOB',

    })

    CONSTRAINT_AUTO_INCREMENT = 'AUTO_INCREMENT'
    STMT_TABLE_TRAILER = 'ENGINE = InnoDB'
    def compiler_class(self):
        return MySQLCompiler

class IterateFetchmanyMixin(object):
    """
    Mixin to cause us to fetch in batches using fetchmany().
    """
    sleep = None
    def __iter__(self):
        fetch = self.fetchmany
        sleep = self.sleep
        batch = fetch()
        while batch:
            for row in batch:
                yield row
            if sleep is not None:
                sleep() # pylint:disable=not-callable
            batch = fetch()

    next = __next__ = None

class AbstractMySQLDriver(AbstractModuleDriver):

    MY_SESSION_VARS = {
        # Don't try to decode pickle states as UTF-8 (or whatever the
        # environment is configured as); See
        # https://github.com/zodb/relstorage/issues/57. This varies
        # depending on Python 2/3 and which driver.
        #
        # In the past, we used 'SET names binary' for everything except
        # mysqlclient on Python 3, which set the character_set_results
        # only (because of an issue decoding column names (See
        # https://github.com/zodb/relstorage/issues/213)). But having
        # binary be the default encoding for string literals prevents
        # using the JSON type. So we took another look at connection
        # parameters and got things working with ``character_set_results``
        # everywhere.
        'character_set_results': 'binary',

        # Make the default timezone UTC. That way UTC_TIMESTAMP()
        # and UNIX_TIMESTAMP() and FROM_UNIXTIME are all self-consistent.
        # Subclasses can set to None if they don't need to do this.
        # Starting in 8.0.17 this is hintable using SET_VAR.
        "time_zone": "'+00:00'",

        # Certain identifiers are quoted in sql strings for compatibility
        # with other databases. This should go away as all plain strings
        # transition to modeled values. (Notable example: "transaction"
        # is a reserved word in sqlite.)
        'sql_mode': "CONCAT(@@sql_mode, ',ANSI_QUOTES')",
    }

    dialect = MySQLDialect()

    #: Holds the first statement that all connections need to run.
    #: Used as the value of the ``init_command`` argument to ``connect()`` for
    #: drivers that support it (mysqlclient, PyMySQL). This implementation fills
    #: in the session variables needed.
    _init_command = ''

    def __init__(self):
        super(AbstractMySQLDriver, self).__init__()

        kv = ["%s=%s" % (k, v) for k, v in self.MY_SESSION_VARS.items()]
        self._init_command = "SET " + ", ".join(kv)

    _server_side_cursor = None

    def cursor(self, conn, server_side=False):
        if server_side:
            assert self._server_side_cursor is not None
            cursor = conn.cursor(self._server_side_cursor)
            cursor.arraysize = self.cursor_arraysize
        else:
            cursor = super(AbstractMySQLDriver, self).cursor(conn, server_side=False)
        return cursor

    def connect(self, *args, **kwargs):
        if self._init_command:
            kwargs['init_command'] = self._init_command
        return super(AbstractMySQLDriver, self).connect(*args, **kwargs)

    def synchronize_cursor_for_rollback(self, cursor):
        """Does nothing."""

    def callproc_multi_result(self, cursor, proc,
                              args=(),
                              exit_critical_phase=False):
        """
        Some drivers need extra arguments to execute a statement that
        returns multiple results, and they don't all use the standard
        way to retrieve them, so use this.

        Returns a list of lists of rows, one list for each result set::

            [
              [(row in first),  ...],
              [(row in second), ...],
              ...
            ]

        The last item in the list is the result of the stored procedure
        itself.

        Note that, because 'CALL' potentially returns multiple result
        sets, there is potentially at least one extra database round
        trip involved when we call ``cursor.nextset()``. If the
        procedure being called is very short or returns only a single
        very small result, this may add substantial overhead.

        As of PyMySQL 0.9.3, mysql-connector-python 8.0.16 and MySQLdb
        (mysqlclient) 1.4.2 using libmysqlclient.21.so (from mysql8)
        or libmysqlclient.20 (mysql 5.7), all the drivers use the
        flags from the server to detect that there are no more results
        and turn ``nextset()`` into a simple flag check. So if the CALL
        only returns one result set (because the CALLed object doesn't
        return any of its own, i.e., it only has side-effects) there
        shouldn't be any penalty.

        :keyword bool exit_critical_phase: If true (*not* the default),
            then if the cursor's connection is in a critical phase,
            it will be exited *before* sending the ``CALL`` query.
            Do this when the called procedure will commit the transaction
            itself. There's no need to maintain elevated priority and block
            other greenlets while the query completes, we have nothing else
            that must be done to release locks and unblock other threads.

            Exiting before is fine; it's almost a certainty that we
            won't need to block to write to the socket.
        """
        if exit_critical_phase:
            self.exit_critical_phase(cursor.connection, cursor)

        cursor.execute('CALL ' + proc, args)

        multi_results = [cursor.fetchall()]
        while cursor.nextset():
            multi_results.append(cursor.fetchall())
        return multi_results

    def callproc_no_result(self, cursor, proc, args=()):
        """
        An optimization for calling stored procedures that don't
        produce any results, except for the null result of calling the
        procedure itself.
        """
        cursor.execute('CALL ' + proc, args)
        cursor.fetchall()

    def exit_critical_phase(self, connection, cursor):
        "Override if you implement critical phases."


implement_db_driver_options(
    __name__,
    'mysqlconnector', 'mysqldb', 'pymysql',
)



"""
Transaction support for Blobs tests
"""
from __future__ import absolute_import
from __future__ import print_function

import os

from ZODB.interfaces import BlobError
from ZODB.interfaces import IStorageUndoable
from ZODB.blob import Blob
from ZODB.POSException import ConflictError
from ZODB.POSException import ConnectionStateError
from ZODB.POSException import POSKeyError

import transaction

from relstorage.blobhelper.interfaces import IAuthoritativeBlobHelper
from relstorage.blobhelper.interfaces import ICachedBlobHelper
from . import TestBlobMixin

class TestBlobTransactionMixin(TestBlobMixin):

    DATA1 = b'this is blob 1'

    def _make_blob(self, data=DATA1):
        blob1 = Blob()
        with blob1.open('w') as file:
            file.write(data)
        return blob1

    def _make_and_commit_blob(self, data=DATA1, close=True):
        """
        Stores a blob at root['blob1']
        """
        blob = self._make_blob(data)
        conn = self.database.open()
        root1 = conn.root()
        root1['blob1'] = blob
        transaction.commit()
        if close:
            conn.close()
            return blob
        return blob, conn

    def _check_blob_committed(self, blob, uncommitted_fname=None):
        self.assertIsNone(blob._p_blob_uncommitted)
        if uncommitted_fname:
            self.assertFalse(os.path.exists(uncommitted_fname), uncommitted_fname)

    def _check_blob_uncommitted(self, blob):
        uncommitted_fname = blob._p_blob_uncommitted
        self.assertTrue(os.path.exists(uncommitted_fname), uncommitted_fname)
        return uncommitted_fname

    def _check_blob_contents(self, blob, expected, mode='r'):
        with blob.open(mode) as fp:
            data = fp.read()
        self.assertEqual(data, expected)

    def test_abort_transaction(self):
        # Putting a Blob into a Connection works like any other Persistent object::
        blob1 = self._make_blob()
        conn = self.database.open()
        root1 = conn.root()
        self.assertNotIn('blob1', root1)
        root1['blob1'] = blob1
        self.assertIn('blob1', root1)
        self.assertEqual(root1._p_status, 'changed')
        # Aborting a blob add leaves the blob unchanged:

        transaction.abort()
        # Note that debugging with something like pudb that automatically
        # displays local variables will invalidate this: it will already
        # have been loaded again.
        self.assertEqual(root1._p_status, 'ghost')
        self.assertNotIn('blob1', root1)

        self.assertIsNone(blob1._p_oid)
        self.assertIsNone(blob1._p_jar)
        self._check_blob_contents(blob1, self.DATA1)

        # It doesn't clear the file because there is no previously committed version:
        uncommitted_fname = self._check_blob_uncommitted(blob1)

        # Let's put the blob back into the root and commit the change:
        root1['blob1'] = blob1
        transaction.commit()

        # Now, if we make a change and abort it, we'll return to the committed
        # state:
        new_data = b'this is new blob 1'
        with blob1.open('w') as file:
            file.write(b'this is new blob 1')
        self._check_blob_contents(blob1, new_data)
        uncommitted_fname = self._check_blob_uncommitted(blob1)

        transaction.abort()
        self._check_blob_committed(blob1, uncommitted_fname)
        self._check_blob_contents(blob1, self.DATA1)
        conn.close()

    def test_multiple_open_handles(self):
        # Opening a blob gives us a filehandle.  Getting data out of the
        # resulting filehandle is accomplished via the filehandle's read method::
        self._make_and_commit_blob()

        connection2 = self.database.open()
        root2 = connection2.root()
        blob1a = root2['blob1']

        blob1afh1 = blob1a.open("r")
        self.assertEqual(blob1afh1.read(), self.DATA1)


        # Let's make another filehandle for read only to blob1a. Each file
        # handle has a reference to the (same) underlying blob::
        blob1afh2 = blob1a.open("r")
        self.assertIs(blob1afh2.blob, blob1afh1.blob)

        # Let's close the first filehandle we got from the blob::
        blob1afh1.close()

        # Let's abort this transaction, and ensure that the other filehandle that we
        # opened are still open::
        transaction.abort()
        self.assertEqual(blob1afh2.read(), self.DATA1)
        blob1afh2.close()

        connection2.close()

    def test_append_makes_dirty(self):
        # If we open a blob for append, writing any number of bytes to the
        # blobfile should result in the blob being marked "dirty" in the
        # connection (we just aborted above, so the object should be "clean"
        # when we start)::
        blob, conn = self._make_and_commit_blob(close=False)
        self.assertFalse(blob._p_changed)
        self._check_blob_contents(blob, self.DATA1)
        with blob.open('a') as blob1afh3:
            self.assertTrue(blob._p_changed)
            blob1afh3.write(b'woot!')

        transaction.abort()
        conn.close()

    def test_multiple_blobs(self):
        # We can open more than one blob object during the course of a single
        # transaction, and we can write to a single blob using
        # multiple handles.
        blob1, conn1 = self._make_and_commit_blob(close=False)
        root = conn1.root()
        with blob1.open('a') as f:
            f.write(b'woot')

        blob1_second_object = root['blob1']
        self.assertEqual(blob1_second_object._p_oid, blob1._p_oid)
        self.assertIs(blob1_second_object, blob1)

        with blob1_second_object.open('a') as f:
            f.write(b'!')

        blob2 = Blob()
        blob2_contents = b'this is blob 2'
        with blob2.open('w') as f:
            f.write(blob2_contents)
            root['blob2'] = blob2
        transaction.commit()

        # Since we committed the current transaction above, the aggregate
        # changes we've made to blob, blob1a (these refer to the same object) and
        # blob2 (a different object) should be evident::
        blob1_contents = self.DATA1 + b'woot!'
        self._check_blob_contents(blob1, blob1_contents)
        self._check_blob_contents(blob1_second_object, blob1_contents)
        self._check_blob_contents(blob2, blob2_contents)

        transaction.abort()
        conn1.close()

        if IStorageUndoable.providedBy(self.blob_storage):
            # Whether or not we're using a shared or unshared
            # blob-dir, when we keep history we will have three blob
            # files on disk: two revisions of blob1, and one revision
            # of blob2
            self.assertEqual(3, self._count_blobs_in_directory())
        else:
            # If we are a shared blob directory, we didn't remove anything;
            # that waits until pack time.
            if IAuthoritativeBlobHelper.providedBy(self.blob_storage.blobhelper):
                self.assertEqual(3, self._count_blobs_in_directory())
            else:
                # We will just have two blobs on disk. The earlier revision
                # was automatically removed.
                self.assertTrue(ICachedBlobHelper.providedBy(self.blob_storage.blobhelper))
                self.assertEqual(2, self._count_blobs_in_directory())

    def test_persistent_blob_handle(self):
        # We shouldn't be able to persist a blob filehandle at commit time
        # (although the exception which is raised when an object cannot be
        # pickled appears to be particulary unhelpful for casual users at the
        # moment)::
        blob1, conn = self._make_and_commit_blob(close=False)
        root = conn.root()
        with blob1.open('r') as f:
            root['wontwork'] = f
            with self.assertRaises(TypeError):
                transaction.commit()

        # Abort for good measure::

        transaction.abort()
        conn.close()

    def test_write_conflict(self):
        # Attempting to change a blob simultaneously from two different
        # connections should result in a write conflict error::
        self._make_and_commit_blob()

        tm1 = transaction.TransactionManager()
        tm2 = transaction.TransactionManager()
        database = self.database
        conn3 = database.open(transaction_manager=tm1)
        conn4 = database.open(transaction_manager=tm2)
        root3 = conn3.root()
        root4 = conn4.root()
        blob1c3 = root3['blob1']
        blob1c4 = root4['blob1']
        conn3_data = b'this is from connection 3'
        conn4_data = b'this is from connection 4'
        with blob1c3.open('w') as blob1c3fh1:
            blob1c3fh1.write(conn3_data)
        with blob1c4.open('w') as blob1c4fh1:
            blob1c4fh1.write(conn4_data)
        tm1.commit()
        self._check_blob_contents(root3['blob1'], conn3_data)
        with self.assertRaises(ConflictError):
            tm2.commit()

        # After the conflict, the winning transaction's result is visible on both
        # connections::
        tm2.abort()
        self._check_blob_contents(root4['blob1'], conn3_data)

        conn3.close()
        conn4.close()

    def test_commit_while_open(self):
        # You can't commit a transaction while blob files are open:
        conn = self.database.open()
        blob = self._make_blob()
        conn.root()['blob1'] = blob
        f = blob.open('w')
        with self.assertRaisesRegex(ValueError, "Can't commit with opened blobs"):
            transaction.commit()

        f.close()
        transaction.abort()

        # Open for reading, if we haven't changed anything, is fine
        f = blob.open('r')
        transaction.commit()
        f.close()

        # But not if we've made a change
        blob.open('w').close()
        f = blob.open('r')
        with self.assertRaisesRegex(ValueError, "Can't commit with opened blobs"):
            transaction.commit()
        f.close()
        transaction.abort()
        conn.close()

    def _list_savepoints(self):
        blob_dir = self.blob_storage.blobhelper.blob_dir
        return [name for name in os.listdir(os.path.join(blob_dir, 'tmp'))
                if name.startswith('savepoint')]

    def test_optimistic_savepoints(self):
        # We do support optimistic savepoints:
        blob, conn = self._make_and_commit_blob(close=False)
        data2 = b" Here's more data."
        with blob.open("a") as blob_fh:
            blob_fh.write(data2)
        self._check_blob_contents(blob, self.DATA1 + data2)

        _ = transaction.savepoint(optimistic=True)
        self._check_blob_contents(blob, self.DATA1 + data2)

        # Savepoints store the blobs in temporary directories in the
        # temporary directory of the blob storage:
        self.assertEqual(
            1,
            len(self._list_savepoints()))
        # After committing the transaction, the temporary savepoint
        # files are moved to the committed location again:

        transaction.commit()
        self.assertEmpty(self._list_savepoints())
        transaction.abort()
        conn.close()

    def test_non_optimistic_savepoints(self):
        # We support non-optimistic savepoints too:

        blob, conn = self._make_and_commit_blob(close=False)
        new_data = b' More data.'
        with blob.open('a') as f:
            f.write(new_data)
        self._check_blob_contents(blob, self.DATA1 + new_data)

        savepoint = transaction.savepoint()

        # Again, the savepoint creates a new savepoints directory:
        self.assertEqual(1, len(self._list_savepoints()))

        with blob.open("w") as f:
            f.write(b"Completely unrelated.")
        savepoint.rollback()

        self._check_blob_contents(blob, self.DATA1 + new_data)
        transaction.abort()

        # The savepoint blob directory gets cleaned up on an abort:
        self.assertEmpty(self._list_savepoints())
        conn.close()

    def test_reading_outside_transaction(self):
        # Reading Blobs outside of a transaction
        # --------------------------------------

        # If you want to read from a Blob outside of transaction boundaries (e.g. to
        # stream a file to the browser), committed method to get the name of a
        # file that can be opened.
        data = b"I'm a happy blob."
        blob = self._make_and_commit_blob(data)

        # Of course, we can't do that if the connection has been completely closed
        with self.assertRaises(ConnectionStateError):
            blob.committed()

        conn = self.database.open()

        path = blob.committed()
        with open(path, 'rb') as fp:
            read = fp.read()
        self.assertEqual(read, data)


        # We can also read committed data by calling open with a 'c' flag:
        self._check_blob_contents(blob, data, 'c')

        # This doesn't prevent us from opening the blob for writing:

        with blob.open('w') as file:
            file.write(b'x')
        self._check_blob_contents(blob, b'x')
        # This doesn't change the committed contents
        with open(path, 'rb') as fp:
            read = fp.read()
        self.assertEqual(read, data)

        # Although we can't use the 'c' flag anymore
        with self.assertRaisesRegex(BlobError, 'Uncommitted changes'):
            self._check_blob_contents(blob, data, 'c')
        transaction.abort()
        conn.close()

    def test_committed_when_uncommitted(self):
        # An exception is raised if we call committed on a blob that has
        # uncommitted changes:
        blob = Blob()
        with self.assertRaisesRegex(BlobError, "Uncommitted changes"):
            blob.committed()

        with self.assertRaisesRegex(BlobError, "Uncommitted changes"):
            blob.open('c')

        blob = self._make_blob()
        conn = self.database.open()
        root = conn.root()
        root['blob'] = blob
        with self.assertRaisesRegex(BlobError, "Uncommitted changes"):
            blob.committed()

        with self.assertRaisesRegex(BlobError, "Uncommitted changes"):
            blob.open('c')

        transaction.savepoint()
        with self.assertRaisesRegex(BlobError, "Uncommitted changes"):
            blob.committed()
        with self.assertRaisesRegex(BlobError, "Uncommitted changes"):
            blob.open('c')

        transaction.commit()
        self._check_blob_contents(blob, self.DATA1, 'c')

    def test_committed_cannot_open_for_write(self):
        # You can't open a committed blob file for writing:
        blob, conn = self._make_and_commit_blob(close=False)
        # Raises IOError on Python2, and PersmissionError on Python 3,
        # but PermissionError extends OSError (which is the new name for IOError)
        with self.assertRaises(IOError):
            open(blob.committed(), 'w')
        conn.close()

    def test_tpc_abort(self):
        # If a transaction is aborted in the middle of 2-phase commit, any data
        # stored are discarded.
        blob_storage = self.blob_storage
        blob, conn = self._make_and_commit_blob(close=False)

        olddata, oldserial = blob_storage.load(blob._p_oid, '')
        t = transaction.get()
        blob_storage.tpc_begin(t)
        with open('blobfile', 'wb') as file:
            file.write(b'This data should go away')
        blob_storage.storeBlob(blob._p_oid, oldserial, olddata, 'blobfile',
                               '', t)
        new_oid = blob_storage.new_oid()
        with open('blobfile2', 'wb') as file:
            file.write(b'This data should go away too')
        blob_storage.storeBlob(new_oid, b'\0' * 8, olddata, 'blobfile2',
                               '', t)
        self.assertFalse(blob_storage.tpc_vote(t))

        blob_storage.tpc_abort(t)

        # Now, the serial for the existing blob should be the same:

        self.assertEqual(blob_storage.load(blob._p_oid, ''),
                         (olddata, oldserial))

        # The old data should be unaffected:
        with open(blob_storage.loadBlob(blob._p_oid, oldserial), 'rb') as fp:
            data = fp.read()
        self.assertEqual(data, self.DATA1)


        # Similarly, the new object wasn't added to the storage:
        with self.assertRaises(POSKeyError):
            blob_storage.load(new_oid, '')

        conn.close()



##############################################################################
#
# Copyright (c) 2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
An implementation of ``IStateCache`` using a memcache client.

Keys and values are transformed into the (byte)string based
keys and values that memcache accepts.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import importlib

from ZODB.utils import p64
from ZODB.utils import u64
from zope import interface

from relstorage._compat import string_types
from relstorage._compat import iteritems
from relstorage.cache.interfaces import IStateCache


@interface.implementer(IStateCache)
class MemcacheStateCache(object):

    # send_limit: approximate limit on the bytes to buffer before
    # sending to the cache.
    send_limit = 1024 * 1024

    @classmethod
    def from_options(cls, options, prefix=''):
        """
        Create and return a MemcacheStateCache from the options,
        if they so request.
        """
        if not options.cache_servers:
            return
        module_name = options.cache_module_name
        module = importlib.import_module(module_name)
        servers = options.cache_servers
        if isinstance(servers, string_types):
            servers = servers.split()

        return cls(
            lambda: module.Client(servers),
            prefix
        )

    def __init__(self, constructor, prefix):
        self._constructor = constructor
        self.prefix = prefix
        self.client = constructor()
        # checkpoints_key holds the current checkpoints.
        self.checkpoints_key = ck = '%s:checkpoints' % self.prefix
        # no unicode on Py2
        assert isinstance(ck, str), (ck, type(ck))

    def __oid_tid_to_key(self, oid, tid):
        return '%s:state:%d:%d' % (self.prefix, tid, oid)

    def __getitem__(self, oid_tid, peek=False):
        oid, tid = oid_tid
        if tid is None:
            # We don't support frozen keys, only those in the index
            return None

        cachekeys = [self.__oid_tid_to_key(oid, tid)]
        response = self.client.get_multi(cachekeys) or {}
        for key in cachekeys:
            data = response.get(key)
            if data and len(data) >= 8:
                actual_tid_int = u64(data[:8])
                return data[8:], actual_tid_int

    get = __getitem__

    def __contains__(self, oid_tid):
        return self[oid_tid] is not None

    def __setitem__(self, oid_tid, state_bytes_tid):
        oid, tid = oid_tid
        key = self.__oid_tid_to_key(oid, tid)
        state_bytes, actual_tid = state_bytes_tid
        cache_data = p64(actual_tid) + (state_bytes or b'')
        self.client.set(key, cache_data)

    def __delitem__(self, oid_tid):
        self.client.delete(self.__oid_tid_to_key(*oid_tid))

    def invalidate_all(self, oids): # pylint:disable=unused-argument
        """
        Implemented by flushing everything.
        """
        self.flush_all()

    def _set_multi(self, keys_and_values):
        formatted = {
            '%s:state:%d:%d' % (self.prefix, tid, oid): (p64(actual_tid) + (state or b''))
            for (oid, tid), (state, actual_tid) in iteritems(keys_and_values)
        }
        self.client.set_multi(formatted)

    def set_all_for_tid(self, tid_int, state_oid_iter):
        send_size = 0
        to_send = {}
        for state, oid_int, _ in state_oid_iter:
            length = len(state)
            cachekey = (oid_int, tid_int)
            item_size = length + len(cachekey)
            if send_size and send_size + item_size >= self.send_limit:
                self._set_multi(to_send)
                to_send.clear()
                send_size = 0
            to_send[cachekey] = (state, tid_int)
            send_size += item_size

        if to_send:
            self._set_multi(to_send)

    def store_checkpoints(self, cp0_tid, cp1_tid):
        checkpoint_data = '%d %d' % (cp0_tid, cp1_tid)
        checkpoint_data = checkpoint_data.encode('ascii')
        self.client.set(self.checkpoints_key, checkpoint_data)

    def get_checkpoints(self):
        s = self.client.get(self.checkpoints_key)
        if s:
            try:
                c0, c1 = s.split()
                c0 = int(c0)
                c1 = int(c1)
            except ValueError:
                # Invalid checkpoint cache value; ignore it.
                pass
            else:
                # More validation
                return (c0, c1) if c0 >= c1 else None

    def replace_checkpoints(self, expected, change_to):
        cp = self.get_checkpoints()
        if cp is not None and cp != expected:
            return None
        self.store_checkpoints(*change_to)
        return change_to

    def close(self):
        if self.client is not None:
            self.client.disconnect_all()
            self.client = None

    release = close

    def flush_all(self):
        self.client.flush_all()

    def updating_delta_map(self, deltas):
        return deltas

    def new_instance(self):
        return type(self)(self._constructor, self.prefix)



# -*- coding: utf-8 -*-
"""
Expressions in the AST.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from zope.interface import implementer

from ._util import Resolvable
from ._util import copy

from .dialect import DialectAware
from .interfaces import INamedBindParam
from .interfaces import IOrderedBindParam

class Expression(DialectAware,
                 Resolvable):
    """
    A SQL expression.
    """

    __slots__ = ()

@implementer(INamedBindParam)
class BindParam(Expression):

    __slots__ = (
        'key',
    )

    def __init__(self, key):
        self.key = key

    def __compile_visit__(self, compiler):
        compiler.visit_bind_param(self)

@implementer(INamedBindParam)
def bindparam(key):
    return BindParam(key)


class LiteralExpression(Expression):

    __slots__ = (
        'value',
    )

    def __init__(self, value):
        self.value = value

    def __str__(self):
        return str(self.value)

    def __compile_visit__(self, compiler):
        compiler.visit_literal_expression(self.value)

class BooleanLiteralExpression(LiteralExpression):

    __slots__ = ()

    def __compile_visit__(self, compiler):
        compiler.visit_boolean_literal_expression(self.value)

@implementer(IOrderedBindParam)
class OrderedBindParam(Expression):

    __slots__ = ()

    name = '%s'

    def __compile_visit__(self, compiler):
        compiler.visit_ordered_bind_param(self)


@implementer(IOrderedBindParam)
def orderedbindparam():
    return OrderedBindParam()


def as_expression(stmt):
    if hasattr(stmt, '__compile_visit__'):
        return stmt

    if isinstance(stmt, bool):
        # The values True and False are handled
        # specially because their representation varies
        # among databases (Oracle)
        stmt = BooleanLiteralExpression(stmt)
    else:
        stmt = LiteralExpression(stmt)
    return stmt

class BinaryExpression(Expression):
    """
    Expresses a comparison.
    """

    __slots__ = (
        'op',
        'lhs',
        'rhs',
    )

    def __init__(self, op, lhs, rhs):
        self.op = op
        self.lhs = lhs # type: Column
        # rhs is either a literal or a column;
        # certain literals are handled specially.
        rhs = as_expression(rhs)
        self.rhs = rhs

    def __str__(self):
        return '%s %s %s' % (
            self.lhs,
            self.op,
            self.rhs
        )

    def _visit_lhs(self, compiler):
        compiler.visit(self.lhs)

    def _visit_rhs(self, compiler):
        compiler.visit(self.rhs)

    def __compile_visit__(self, compiler):
        self._visit_lhs(compiler)
        compiler.visit_op(self.op)
        self._visit_rhs(compiler)

    def resolve_against(self, table):
        lhs = self.lhs.resolve_against(table)
        rhs = self.rhs.resolve_against(table)
        new = copy(self)
        new.rhs = rhs
        new.lhs = lhs
        return new

class EmptyExpression(Expression):
    """
    No expression at all.
    """

    __slots__ = ()

    def __bool__(self):
        return False

    __nonzero__ = __bool__

    def __str__(self):
        return ''

    def and_(self, expression):
        return expression

    def __compile_visit__(self, compiler):
        "Does nothing"

class EqualExpression(BinaryExpression):

    __slots__ = ()

    def __init__(self, lhs, rhs):
        BinaryExpression.__init__(self, '=', lhs, rhs)

AssignmentExpression = EqualExpression

class NotEqualExpression(BinaryExpression):

    __slots__ = ()

    def __init__(self, lhs, rhs):
        BinaryExpression.__init__(self, '<>', lhs, rhs)


class GreaterExpression(BinaryExpression):

    __slots__ = ()

    def __init__(self, lhs, rhs):
        BinaryExpression.__init__(self, '>', lhs, rhs)

class GreaterEqualExpression(BinaryExpression):

    __slots__ = ()

    def __init__(self, lhs, rhs):
        BinaryExpression.__init__(self, '>=', lhs, rhs)

class LessEqualExpression(BinaryExpression):

    __slots__ = ()

    def __init__(self, lhs, rhs):
        BinaryExpression.__init__(self, '<=', lhs, rhs)

class LessExpression(BinaryExpression):

    __slots__ = ()

    def __init__(self, lhs, rhs):
        BinaryExpression.__init__(self, '<', lhs, rhs)

class InExpression(BinaryExpression):
    __slots__ = ()

    def __init__(self, lhs, rhs):
        BinaryExpression.__init__(self, 'IN', lhs, rhs)

    def _visit_rhs(self, compiler):
        compiler.visit_grouped(self.rhs)


class And(Expression):

    __slots__ = (
        'lhs',
        'rhs',
    )

    def __init__(self, lhs, rhs):
        self.lhs = as_expression(lhs)
        self.rhs = as_expression(rhs)

    def __compile_visit__(self, compiler):
        compiler.visit_grouped(BinaryExpression('AND', self.lhs, self.rhs))

    def resolve_against(self, table):
        lhs = self.lhs.resolve_against(table)
        rhs = self.rhs.resolve_against(table)
        new = copy(self)
        self.lhs = lhs
        self.rhs = rhs
        return new


class ParamMixin(object):
    def bindparam(self, key):
        return bindparam(key)

    def orderedbindparam(self):
        return orderedbindparam()

class ExpressionOperatorMixin(object):
    def __eq__(self, other):
        return EqualExpression(self, other)

    def __gt__(self, other):
        return GreaterExpression(self, other)

    def __ge__(self, other):
        return GreaterEqualExpression(self, other)

    def __ne__(self, other):
        return NotEqualExpression(self, other)

    def __le__(self, other):
        return LessEqualExpression(self, other)

    def __lt__(self, other):
        return LessExpression(self, other)

    def in_(self, other):
        return InExpression(self, other)



##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
Database schema installers
"""
from __future__ import absolute_import

from collections import namedtuple

from ZODB.POSException import StorageError
from zope.interface import implementer

from ..connmanager import connection_callback
from ..interfaces import ISchemaInstaller
from ..schema import AbstractSchemaInstaller
from ..schema import Schema
from .._util import DatabaseHelpersMixin

logger = __import__('logging').getLogger(__name__)

###
# Notes on procedures vs functions.
#
# When we have a result, we'd generally like to use a function,
# and simply ``SELECT func()``.
#
# However, MySQL is badly broken and functions play poorly with
# transaction logging (replication logging). For that reason, to
# create a function, you must have SUPER privileges (essentially root)
# so we can't count on being able to do that.
# (https://dev.mysql.com/doc/refman/5.7/en/stored-programs-logging.html).
# (Also, functions take out table locks and limit concurrency, and
# must be marked DETERMINISTIC or NO SQL or just READS SQL DATA ---
# they can't make modifications.)
#
# So we're stuck with procedures.
#
# The procedure should be called with ``cursor.execute('CALL
# proc_name(%s)')`` instead of ``cursor.callproc()`` ---
# ``callproc()`` involves setting server variables in an extra trip to
# the server. That would enable us to use an OUT param for the result,
# but getting that would be another round trip too. Procedures that
# generate results via a SELECT statement also generate a blank result
# that must be retrieved with cursor.nextset().
# ``Driver.callproc_multi_result`` handles the details.


# NOTE: Unlike PostgreSQL, NOW() and UTC_TIMESTAMP() are only consistent
# within a single *statement*; that is, unlike PostgreSQL, these values
# can change within a transaction.

_StoredProcedure = namedtuple('_StoredProcedure',
                              'checksum character_set_client collation_connection')


@implementer(ISchemaInstaller)
class MySQLSchemaInstaller(AbstractSchemaInstaller):

    database_type = 'mysql'
    COLTYPE_BINARY_STRING = 'BLOB'
    TRANSACTIONAL_TABLE_SUFFIX = 'ENGINE = InnoDB'
    COLTYPE_MD5 = 'CHAR(32) CHARACTER SET ascii'
    COLTYPE_STATE = 'LONGBLOB'
    COLTYPE_BLOB_CHUNK = 'LONGBLOB'


    # The names of tables that in the past were explicitly declared as
    # MyISAM but which should now be InnoDB to work with transactions.
    tables_that_used_to_be_myisam_should_be_innodb = (
        'new_oid',
        'object_ref',
        'object_refs_added',
        'pack_object',
        'pack_state',
        'pack_state_tid',
    )

    _PROCEDURES = {}

    def __init__(self, driver=None, version_detector=None, **kwargs):
        self.driver = driver
        self.version_detector = version_detector or MySQLVersionDetector()
        super(MySQLSchemaInstaller, self).__init__(**kwargs)

    def get_database_name(self, cursor):
        cursor.execute("SELECT DATABASE()")
        for (name,) in cursor:
            return self._metadata_to_native_str(name)

    def list_procedures(self, cursor):
        cursor.execute("SHOW PROCEDURE STATUS WHERE db = database()")
        native = self._metadata_to_native_str
        return {
            native(row['name']): _StoredProcedure(
                native(row['comment']),
                native(row['character_set_client']),
                native(row['collation_connection']))
            for row in self._rows_as_dicts(cursor)
        }

    def list_tables(self, cursor):
        return list(self.__list_tables_and_engines(cursor))

    def __list_tables_and_engines(self, cursor):
        # {table_name: engine}, all in lower case.
        cursor.execute("SHOW  TABLE STATUS")
        native = self._metadata_to_native_str

        result = {
            native(row['name']): native(row['engine']).lower()
            for row in self._rows_as_dicts(cursor)
            # This also returns views for some reason, but they don't have
            # an engine.
            if row['engine']
        }
        return result

    def __list_tables_not_innodb(self, cursor):
        return {
            k: v
            for k, v in self.__list_tables_and_engines(cursor).items()
            if k in self.all_tables and v != 'innodb'
        }

    def list_sequences(self, cursor):
        return []

    def list_views(self, cursor):
        cursor.execute("SHOW FULL TABLES WHERE TABLE_TYPE LIKE 'VIEW'")
        return [self._metadata_to_native_str(r[0]) for r in cursor]

    def check_compatibility(self, cursor, tables):
        super(MySQLSchemaInstaller, self).check_compatibility(cursor, tables)
        tables_that_are_not_innodb = self.__list_tables_not_innodb(cursor)
        if tables_that_are_not_innodb:
            raise StorageError(
                "All RelStorage tables should be InnoDB; MyISAM is no longer supported. "
                "These tables are not using InnoDB: %r" % (tables_that_are_not_innodb,)
            )

    def _create_pack_lock(self, cursor):
        return

    # As usual, MySQL has a quirky implementation of this feature and we
    # have to re-specify *everything* about the column. MySQL 8 supports the
    # simple 'RENAME ... TO ...' syntax that everyone else does.
    _rename_transaction_empty_stmt = (
        "ALTER TABLE transaction CHANGE empty is_empty "
        "BOOLEAN NOT NULL DEFAULT FALSE"
    )

    _new_oid_query = Schema.new_oid.create()

    def _create_new_oid(self, cursor):
        self._new_oid_query.execute(cursor)

    # Temp tables are created in a session-by-session basis
    def _create_temp_store(self, _cursor):
        return

    def _create_temp_blob_chunk(self, _cursor):
        return

    def _create_temp_pack_visit(self, _cursor):
        return

    def _create_temp_undo(self, _cursor):
        return

    def _reset_oid(self, cursor):
        from .oidallocator import MySQLOIDAllocator
        MySQLOIDAllocator(self.driver).reset_oid(cursor)

    def __convert_all_tables_to_innodb(self, cursor):
        tables = self.__list_tables_not_innodb(cursor)
        if not tables:
            logger.debug("All tables already InnoDB")
            return

        logger.info("Converting tables to InnoDB: %s", tables)
        for table in tables:
            logger.info("Converting table %s to Innodb", table)
            cursor.execute("ALTER TABLE %s ENGINE=Innodb" % (table,))
        logger.info("Done converting tables to InnoDB: %s", tables)

    @connection_callback(inherit=AbstractSchemaInstaller._prepare_with_connection)
    def _prepare_with_connection(self, conn, cursor):
        from .oidallocator import MySQLOIDAllocator
        self.__convert_all_tables_to_innodb(cursor)
        super(MySQLSchemaInstaller, self)._prepare_with_connection(conn, cursor)
        MySQLOIDAllocator(self.driver).garbage_collect_oids(cursor)

    def _read_proc_files(self):
        name_to_source = super(MySQLSchemaInstaller, self)._read_proc_files()

        for proc_name, source in name_to_source.items():
            # No leading or trailing lines allowed, only the procedure
            # definition. That way everything is part of the checksum.
            assert source.startswith('CREATE') and source.endswith('END;')
            # Ensure we're creating what we think we are.
            assert proc_name in source
        return name_to_source

    def create_procedures(self, cursor):
        # Apparently procedures remember the ``character_set_client`` and ``collation_connection``
        # that was in use at the time they were defined, and use that
        # to perform implicit conversions on arguments and even internally. If we have
        # done ``SET NAMES binary`` (as we do on Python 2) and we try to pass a JSON
        # argument, or even parse a string into JSON inside the procedure, it will fail,
        # saying it can't convert binary into JSON. Therefore we must be sure to have
        # an appropriate value for both of those installed here.
        installed = self.list_procedures(cursor)
        current_object = 'current_object' if self.keep_history else 'object_state'
        if self.keep_history:
            object_state_join = """
            INNER JOIN object_state ON (object_state.zoid = cur.zoid
                                       AND object_state.tid = cur.tid)
            """
            object_state_name = 'object_state'
        else:
            object_state_join = ""
            object_state_name = 'cur'

        major_version = self.version_detector.get_major_version(cursor)
        if self.version_detector.supports_nowait(cursor):
            set_lock_timeout = ''
            for_share = 'FOR SHARE NOWAIT'
        else:
            set_lock_timeout = 'SET innodb_lock_wait_timeout = 1;'
            for_share = 'LOCK IN SHARE MODE'

        cursor.execute('SET SESSION character_set_client = utf8, '
                       'collation_connection = utf8_general_ci')

        for name, create_stmt in self.procedures.items():
            __traceback_info__ = name
            checksum = self._checksum_for_str(create_stmt) + '; ver ' + str(major_version)
            create_stmt = create_stmt.format(
                CHECKSUM=checksum,
                CURRENT_OBJECT=current_object,
                OBJECT_STATE_NAME=object_state_name,
                OBJECT_STATE_JOIN=object_state_join,
                SET_LOCK_TIMEOUT=set_lock_timeout,
                FOR_SHARE=for_share
            )
            assert checksum in create_stmt
            if name in installed:
                installed_proc = installed[name]
                stored_checksum = installed_proc.checksum
                character_set_client = installed_proc.character_set_client
                collation_connection = installed_proc.collation_connection
                expected = (checksum, 'utf8', 'utf8_general_ci')
                if expected != (stored_checksum, character_set_client, collation_connection):
                    logger.info(
                        "Re-creating procedure %s due to mismatch %s != %s",
                        name,
                        installed_proc, expected
                    )
                    cursor.execute('DROP PROCEDURE %s' % (name,))
                    del installed[name]
                else:
                    logger.debug(
                        "Checksum for procedure %s matches: %s",
                        name, checksum
                    )

            if name not in installed:
                cursor.execute(create_stmt)

    # We can't TRUNCATE tables that have foreign-key relationships
    # with other tables, but we can drop them. This has to be followed up by
    # creating them again.
    _zap_all_tbl_stmt = 'DROP TABLE %s'

    def _after_zap_all_tables(self, cursor, slow=False):
        if not slow:
            logger.debug("Creating tables after drop")
            self.create_tables(cursor)
            logger.debug("Done creating tables after drop")
        else:
            super(MySQLSchemaInstaller, self)._after_zap_all_tables(cursor, slow)


class MySQLVersionDetector(DatabaseHelpersMixin):

    _version = None
    _version_info = None
    _major_version = None

    def _fetch_version_from_server(self, cursor):
        # The newer alternative to VERSION() isn't available everywhere
        # yet.
        # Hook for testing.
        cursor.execute('SELECT version()')
        ver = cursor.fetchone()[0]
        # PyMySQL on Win/Py3 returns this as a byte string; everywhere
        # else it's native.
        return self._metadata_to_native_str(ver)

    def _setup(self, cursor):
        if self._major_version:
            return

        ver = self._version = self._fetch_version_from_server(cursor)
        # . separated parts, with the last one optionally given a suffix
        ver_parts = ver.split('.')
        ver_parts = [part if '-' not in part else part[:part.index('-')]
                     for part in ver_parts]
        self._version_info = tuple(int(i) for i in ver_parts)
        self._major_version = int(ver[0])

    def get_major_version(self, cursor):
        self._setup(cursor)
        return self._major_version

    def get_version(self, cursor):
        self._setup(cursor)
        return self._version

    def get_version_info(self, cursor):
        self._setup(cursor)
        return self._version_info

    def supports_nowait(self, cursor):
        """
        Can we use ``FOR SHARE NOWAIT``?
        """
        return self.get_major_version(cursor) >= 8

    def supports_transaction_isolation(self, cursor):
        """
        The system variable @@transaction_isolation was added in 5.7.20
        along with @@transaction_read_only. Before that there was @@tx_isolation,
        but that's been removed in MySQL 8.
        """
        return self.get_version_info(cursor) >= (5, 7, 20)

    def supports_good_stored_procs(self, cursor):
        """
        Versions of MySQL prior to 5.7.19 crash when we call the stored procedure.
        See https://github.com/zodb/relstorage/pull/287#issuecomment-515518727
        """
        return self.get_version_info(cursor) >= (5, 7, 19)



##############################################################################
#
# Copyright (c) 2005,2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
Import/export support for blob data
"""
from __future__ import absolute_import
from __future__ import print_function

from ZODB.DB import DB
import ZODB.blob
import transaction
from persistent.mapping import PersistentMapping

from . import TestBlobMixin

class TestBlobImportExportMixin(TestBlobMixin):

    def setUp(self):
        super(TestBlobImportExportMixin, self).setUp()
        self.database1 = self.database
        self.storage2 = self.create_storage('2')
        self.database2 = DB(self.storage2)

    def tearDown(self):
        self.database2.close()
        self.storage2.close()
        super(TestBlobImportExportMixin, self).tearDown()

    def testExport(self):
        # Create our root object for database1:

        connection1 = self.database1.open()
        root1 = connection1.root()

        # Put a couple blob objects in our database1 and on the filesystem:
        transaction.begin()
        self.data1 = b'x'*100000
        blob1 = ZODB.blob.Blob()
        with blob1.open('w') as f:
            f.write(self.data1)
        self.data2 = b'y'*100000
        blob2 = ZODB.blob.Blob()
        with blob2.open('w') as f:
            f.write(self.data2)
        d = PersistentMapping({'blob1':blob1, 'blob2':blob2})
        root1['blobdata'] = d
        transaction.commit()
        connection1.close()

        # Export our blobs from a database1 connection:
        conn = self.database1.open()
        root = conn.root()
        oid = root['blobdata']._p_oid
        exportfile = 'export'
        conn.exportFile(oid, exportfile).close()
        conn.close()
        return exportfile

    def testImport(self):
        # Import our exported data into database2:
        exportfile = self.testExport()
        connection2 = self.database2.open()
        root2 = connection2.root()
        transaction.begin()
        data = root2._p_jar.importFile(exportfile)
        root2['blobdata'] = data
        transaction.commit()

        # Make sure our data exists:
        self.assertEqual({'blob1', 'blob2'},
                         set(root2['blobdata'].keys()))

        with root2['blobdata']['blob1'].open() as f:
            b2d = f.read()
        self.assertEqual(b2d, self.data1)
        with root2['blobdata']['blob2'].open() as f:
            b2d = f.read()
        self.assertEqual(b2d, self.data2)
        transaction.get().abort()



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

"""
Temporary local storage for pickles that
will be uploaded to the database and the local
and memcache.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from tempfile import SpooledTemporaryFile

from relstorage._compat import OID_OBJECT_MAP_TYPE as OidObjectMap
from relstorage._compat import OidObjectMap_max_key
from relstorage._compat import iteroiditems

class TemporaryStorage(object):
    __slots__ = (
        '_queue',
        '_queue_contents',
    )

    def __init__(self):
        # start with a fresh in-memory buffer instead of reusing one that might
        # already be spooled to disk.
        # TODO: An alternate idea would be a temporary sqlite database.
        self._queue = SpooledTemporaryFile(max_size=10 * 1024 * 1024)
        # {oid: (startpos, endpos, prev_tid_int)}
        self._queue_contents = OidObjectMap()

    def reset(self):
        self._queue_contents.clear()
        self._queue.seek(0)

    def store_temp(self, oid_int, state, prev_tid_int=0):
        """
        Queue an object for caching.

        Typically, we can't actually cache the object yet, because its
        transaction ID is not yet chosen.
        """
        queue = self._queue
        queue.seek(0, 2)  # seek to end
        startpos = queue.tell()
        queue.write(state)
        endpos = queue.tell()
        self._queue_contents[oid_int] = (startpos, endpos, prev_tid_int)

    def __len__(self):
        # How many distinct OIDs have been stored?
        # This also lets us be used in a boolean context to see
        # if we've actually stored anything or are closed.
        return len(self._queue_contents)

    @property
    def stored_oids(self):
        return self._queue_contents

    @property
    def max_stored_oid(self):
        return OidObjectMap_max_key(self._queue_contents)

    def _read_temp_state(self, startpos, endpos):
        self._queue.seek(startpos)
        length = endpos - startpos
        state = self._queue.read(length)
        if len(state) != length:
            raise AssertionError("Queued cache data is truncated")
        return state

    def read_temp(self, oid_int):
        """
        Return the bytes for a previously stored temporary item.
        """
        startpos, endpos, _ = self._queue_contents[oid_int]
        return self._read_temp_state(startpos, endpos)

    def __iter__(self):
        return self.iter_for_oids(None)

    def iter_for_oids(self, oids):
        read_temp_state = self._read_temp_state
        for startpos, endpos, oid_int, prev_tid_int in self.items(oids):
            state = read_temp_state(startpos, endpos)
            yield state, oid_int, prev_tid_int

    def items(self, oids=None):
        # Order the queue by file position, which should help
        # if the file is large and needs to be read
        # sequentially from disk.
        items = [
            (startpos, endpos, oid_int, prev_tid_int)
            for (oid_int, (startpos, endpos, prev_tid_int)) in iteroiditems(self._queue_contents)
            if oids is None or oid_int in oids
        ]
        items.sort()
        return items

    def close(self):
        if self._queue is not None:
            self._queue.close()
            self._queue = None
            self._queue_contents = () # Not None so len() keeps working



##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
Locker implementations.
"""

from __future__ import absolute_import

from zope.interface import implementer

from ..interfaces import ILocker
from ..interfaces import UnableToAcquirePackUndoLockError
from ..locker import AbstractLocker


@implementer(ILocker)
class PostgreSQLLocker(AbstractLocker):

    def _on_store_opened_set_row_lock_timeout(self, cursor, restart=False):
        # This only lasts beyond the current transaction if it
        # commits. If we have a rollback, then our effect is lost.
        # Luckily, right here is a fine time to commit.
        self._set_row_lock_timeout(cursor, self.commit_lock_timeout)
        cursor.connection.commit()

    def _set_row_lock_timeout(self, cursor, timeout):
        # This will rollback if the transaction rolls back,
        # which should restart our store.
        # Maybe for timeout == 0, we should do SET LOCAL, which
        # is guaranteed not to persist?
        # Recall postgresql uses milliseconds, but our configuration is given
        # in integer seconds.
        # For tests, though, we accept floating point numbers
        timeout_ms = int(timeout * 1000.0)
        self.driver.set_lock_timeout(cursor, timeout_ms)


    def release_commit_lock(self, cursor):
        # no action needed, locks released with transaction.
        pass

    def _get_commit_lock_debug_info(self, cursor, was_failure=False):
        # When we're called, the transaction is guaranteed to be
        # aborted, so to be able to execute any queries we need to
        # rollback. Unfortunately, that would release the locks that
        # we're holding, which have probably(?) already have been released
        # anyway. The output can be excessive.
        if was_failure:
            cursor.connection.rollback()
            return 'Transaction failed; no lock info available'

        from . import debug_locks
        debug_locks(cursor)
        return self._rows_as_pretty_string(cursor)

    def hold_pack_lock(self, cursor):
        """Try to acquire the pack lock.

        Raise an exception if packing or undo is already in progress.
        """
        cursor.execute("SELECT pg_try_advisory_lock(1)")
        locked = cursor.fetchone()[0]
        if not locked:
            raise UnableToAcquirePackUndoLockError('A pack or undo operation is in progress')

    def release_pack_lock(self, cursor):
        """Release the pack lock."""
        cursor.execute("SELECT pg_advisory_unlock(1)")



#!/usr/bin/env python
from __future__ import print_function
import os
import glob
from os.path import join, dirname, abspath, basename
import warnings

import relstorage

warnings.simplefilter('ignore', DeprecationWarning)


# do not generate .rst for the following modules
SKIP = []


template = '''.. AUTOGENERATED -- will be overwritten (remove this comment to save changes)

%(title)s
%(title_underline)s

.. automodule:: relstorage.%(module)s
    :members:
    :undoc-members:
'''

directory = dirname(abspath(relstorage.__file__))
print('Imported relstorage from %s' % (directory, ))
modules = glob.glob(join(directory, '*.py'))
modules = set(basename(filename).split('.')[0] for filename in modules)
adap_mods = {'adapters.' + basename(x).split('.')[0]
             for x in glob.glob(join(directory, 'adapters', '*.py'))
             if not basename(x).startswith("_")}
modules = modules.union(adap_mods)

for driver in 'mysql', 'oracle', 'postgresql', 'sqlite':
    adap_mods = {'adapters.' + driver + '.' + basename(x).split('.')[0]
                 for x in glob.glob(join(directory, 'adapters', driver, '*.py'))
                 if not basename(x).startswith("_")}
    modules = modules.union(adap_mods)

cache_mods = {'cache.' + basename(x).split('.')[0]
              for x in glob.glob(join(directory, 'cache', '*.py'))
              if not basename(x).startswith("_")}

modules = modules.union(cache_mods)


modules = set(name for name in modules
              if not name.startswith('_'))



def _read(fname, count):
    with open(fname) as f:
        return f.read(count)


def generate_rst_for_module(module, do=True):
    doc_name = 'relstorage.%s' % module
    rst_filename = doc_name + '.rst'
    exists = os.path.exists(rst_filename)
    if exists:
        autogenerated = 'autogenerated' in _read(rst_filename, 200).lower()
        if not autogenerated:
            return doc_name

    if '.' in module:
        m = __import__('relstorage.%s' % module, fromlist=['chicken'])
    else:
        m = __import__('relstorage.%s' % module)
        m = getattr(m, module)
    title = getattr(m, '__doc__', None)
    if title:
        lines = title.strip().splitlines()
        for line in lines:
            # skip leading blanks. Support both styles of docstrings.
            if line:
                title = line.strip()
                break
        title = title.strip(' .')
    prefix = ':mod:`relstorage.%s`' % module

    if title:
        title = prefix + ' -- %s' % (title, )
    else:
        title = prefix
    title_underline = '=' * len(title) # pylint:disable=possibly-unused-variable

    params = globals().copy()
    params.update(locals())
    result = template % params
    if getattr(m, '_no_undoc_members', True):
        result = '\n'.join(result.splitlines()[:-1])

    if exists:
        if _read(rst_filename, len(result) + 1) == result:
            return doc_name # already exists one which is the same

    if do:
        print('Generated %s from %s' % (rst_filename, m.__file__))
        with open(rst_filename, 'w') as f:
            f.write(result)
    else:
        print('Would generate %s from %s' % (rst_filename, m.__file__))
    return doc_name


index_template = '''\
================
Internal Details
================

.. caution:: This document is a reference for exploring the code base and
             tracing errors. The details are subject to change.

.. toctree::

'''

def generate_rst(do=True):
    assert os.path.exists('contents.rst'), 'Wrong directory, contents.rst not found'
    documents = []
    for module in modules:
        if module not in SKIP:
            try:
                document = generate_rst_for_module(module, do=do)
            except AttributeError as e:
                print("Failed to generate rst for", module, e, "; skipping")
            else:
                documents.append(document)

    documents.sort()
    with open('internals.rst', 'w') as f:
        f.write(index_template)
        for doc in documents:
            f.write('   ' + doc + '\n')

    return documents


def iter_autogenerated():
    for module in sorted(modules):
        rst_filename = 'relstorage.%s.rst' % module
        exists = os.path.exists(rst_filename)
        if exists:
            autogenerated = 'autogenerated' in open(rst_filename).read(200).lower()
            if autogenerated:
                yield rst_filename


if __name__ == '__main__':
    import sys
    if sys.argv[1:] == ['show']:
        for filename in iter_autogenerated():
            print(filename)
    elif sys.argv[1:] == ['delete']:
        for filename in iter_autogenerated():
            print('Removing', filename)
            os.unlink(filename)
    elif sys.argv[1:] == ['generate']:
        generate_rst()
    elif sys.argv[1:] == []:
        generate_rst(do=False)
    else:
        sys.exit('Invalid command line: %s' % (sys.argv[1:], ))



##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

import unittest


class ZODBPackScriptTests(unittest.TestCase):

    def setUp(self):
        import os
        import tempfile

        fd, self.db_fn = tempfile.mkstemp()
        os.close(fd)

        cfg = """
        <filestorage>
            path %s
        </filestorage>
        """ % self.db_fn

        fd, self.cfg_fn = tempfile.mkstemp()
        os.write(fd, cfg.encode('ascii'))
        os.close(fd)

    def tearDown(self):
        import os
        if os.path.exists(self.db_fn):
            os.remove(self.db_fn)
        if os.path.exists(self.cfg_fn):
            os.remove(self.cfg_fn)

    def test_pack_with_0_day(self):
        from ZODB.DB import DB
        from ZODB.FileStorage import FileStorage
        from ZODB.POSException import POSKeyError
        import time
        import transaction
        from relstorage.zodbpack import main

        storage = FileStorage(self.db_fn, create=True)
        db = DB(storage)
        conn = db.open()
        conn.root()['x'] = 1
        transaction.commit()
        oid = b'\0' * 8
        state, serial = storage.load(oid, '')
        time.sleep(0.1)
        conn.root()['x'] = 2
        transaction.commit()
        conn.close()
        self.assertEqual(state, storage.loadSerial(oid, serial))
        db.close()
        storage = None

        main(['', '--days=0', self.cfg_fn])

        # packing should have removed the old state.
        storage = FileStorage(self.db_fn)
        self.assertRaises(POSKeyError, storage.loadSerial, oid, serial)
        storage.close()

    def test_pack_defaults(self):
        from ZODB.DB import DB
        from ZODB.FileStorage import FileStorage
        import time
        import transaction
        from relstorage.zodbpack import main

        storage = FileStorage(self.db_fn, create=True)
        db = DB(storage)
        conn = db.open()
        conn.root()['x'] = 1
        transaction.commit()
        oid = b'\0' * 8
        state, serial = storage.load(oid, '')
        time.sleep(0.1)
        conn.root()['x'] = 2
        transaction.commit()
        conn.close()
        self.assertEqual(state, storage.loadSerial(oid, serial))
        db.close()
        storage = None

        main(['', '--days=1', self.cfg_fn])

        # packing should not have removed the old state.
        storage = FileStorage(self.db_fn)
        self.assertEqual(state, storage.loadSerial(oid, serial))
        storage.close()


def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(ZODBPackScriptTests))
    return suite

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



"""
Caching of blob data.

(This test is adapted from ZEO/tests/zeo_blob_cache.test in ZODB3.)

We support 2 modes for providing clients access to blob data:

shared
    Blob data are shared via a network file system.  The client shares
    a common blob directory with the server.

non-shared
    Blob data are loaded from the database and cached locally.
    A maximum size for the blob data can be set and data are removed
    when the size is exceeded.

This test is only for the non-shared (aka, "cached") case.

"""
from __future__ import absolute_import
from __future__ import print_function

import threading

import random2

from ZODB.blob import Blob
import transaction

from relstorage._compat import WIN

from relstorage.tests.util import RUNNING_ON_CI
from . import TestBlobMixin

class TestBlobCacheMixin(TestBlobMixin):

    # Here, we passed a blob_cache_size parameter, which specifies a
    # target blob cache size. This is not a hard limit, but rather a
    # target. It defaults to no limit. We also passed a
    # blob_cache_size_check option. The blob_cache_size_check option
    # specifies the number of bytes, as a percent of the target that
    # can be written or downloaded from the server before the cache
    # size is checked. The blob_cache_size_check option defaults to
    # 100. We passed 10, to check after writing 10% of the target
    # size.
    DEFAULT_BLOB_STORAGE_KWARGS = {
        'blob_cache_size': 3000,
        'blob_cache_size_check': 10
    }

    # Too many client threads really slows us down because of the GIL,
    # but we do need some concurrency.
    CLIENT_COUNT = 4 if not RUNNING_ON_CI else 2

    def setUp(self):
        # and don't actually use native threads in gevent mode. They can't be
        # waited on which makes this test racy.

        from relstorage.blobhelper import cached
        from relstorage._util import thread_spawn
        assert cached.native_thread_spawn is not thread_spawn
        cached.native_thread_spawn = thread_spawn

        super(TestBlobCacheMixin, self).setUp()
        # We're going to wait for any threads we started to finish, so...
        self._old_threads = list(threading.enumerate())

    def _wait_for_all_spawned_threads_to_finish(self):
        # pylint:disable=method-hidden

        # Do this only once, at the end.
        to_join = set(threading.enumerate()) - set(self._old_threads)
        for t in to_join:
            t.join(10)
        self._wait_for_all_spawned_threads_to_finish = lambda: None

    def tearDown(self):
        # # Let the shrink run as many more times as it needs to, if it's waiting.
        from relstorage.blobhelper import cached
        from relstorage._util import spawn as native_thread_spawn
        cached.native_thread_spawn = native_thread_spawn

        self._wait_for_all_spawned_threads_to_finish()
        self._old_threads = []
        super(TestBlobCacheMixin, self).tearDown()

    # Set this and BLOB_SIZE to create a total of 10,000
    # bytes of data
    BLOB_COUNT = 100
    BLOB_SIZE = 100

    def _blob_numbers(self):
        return range(1, 1 + self.BLOB_COUNT)

    def _data_for_blob_number(self, i):
        return (str(i) * self.BLOB_SIZE).encode('ascii')

    def _populate(self):
        conn = self.database.open()
        for i in self._blob_numbers():
            conn.root()[i] = Blob()
            with conn.root()[i].open('w') as f:
                f.write(self._data_for_blob_number(i))
                f.flush()
        transaction.commit()
        conn.close()

    def _verify_blob_number(self, i, conn, mode='r'):
        data_for_blob = self._data_for_blob_number(i)

        blob = conn.root()[i]
        with blob.open(mode) as f:
            read = f.read()
        self.assertEqual(
            read,
            data_for_blob)

    def _verify_all_blobs(self, mode='r'):
        conn = self.database.open()
        for blob_number in self._blob_numbers():
            self._verify_blob_number(blob_number, conn, mode)
        conn.close()

    def _wait_for_shrinks_to_finish(self):
        cache_checker = self.blob_storage.blobhelper.cache_checker
        cache_checker.wait_for_checker()
        size = self._size_blobs_in_directory()
        self.assertLess(size, 5000)

    def test_exceed_size_and_shrink(self):
        self._populate()
        # We've committed 10000 bytes of data, but our target size is 3000.  We
        # expect to have not much more than the target size in the cache blob
        # directory. At the end of the process. Concurrently, cleanups
        # will be going on, racing with us to read.

        # If we read all of the blobs, data will be downloaded again, as
        # necessary, but the cache size will remain not much bigger than the
        # target:
        self._verify_all_blobs()
        self._verify_all_blobs()

        # This time using the 'committed' mode
        self._verify_all_blobs('c')
        self._wait_for_shrinks_to_finish()


    def test_many_clients(self):
        # Now let see if we can stress things a bit.  We'll create many clients
        # and get them to pound on the blobs all at once to see if we can
        # provoke problems:
        self._populate()

        verification_errors = []

        def verify_a_blob(i, conn, mode):
            try:
                self._verify_blob_number(i, conn, mode)
            except Exception as e: # pylint:disable=broad-except
                # relying on the GIL here.
                verification_errors.append(e)

        def client(client_num):
            # Deterministic random numbers so we can track down any failures.
            random = random2.Random(client_num)
            conn = self.database.open()
            try:
                for i in range(300):
                    i = random.randint(1, 100)
                    verify_a_blob(i, conn, 'r')
                    verify_a_blob(i, conn, 'c')
            finally:
                conn.close()


        threads = [threading.Thread(target=client, args=(i,))
                   for i in range(self.CLIENT_COUNT)]
        for thread in threads:
            thread.setDaemon(True)
            thread.start()

        for thread in threads:
            thread.join(10)

        del threads

        self._wait_for_shrinks_to_finish()
        __traceback_info__ = verification_errors
        self.assertNoVerificationErrors(verification_errors)


    def assertNoVerificationErrors(self, error_list):
        self.assertEmpty(error_list)


class TestBlobCacheExternalCleanupMixin(TestBlobCacheMixin):

    DEFAULT_BLOB_STORAGE_KWARGS = dict(TestBlobCacheMixin.DEFAULT_BLOB_STORAGE_KWARGS)
    DEFAULT_BLOB_STORAGE_KWARGS['blob_cache_size_check_external'] = True

    if WIN:
        def assertNoVerificationErrors(self, error_list):
            """
            On Windows, with external cleanups going on, it's possible
            to get ``WindowsError(32, 'The process cannot access the
            file because it is being used by another process')`` as the two
            race each other to check file sizes.

            We don't recommend external cleanups on Windows, so this
            does no verification.
            """



##############################################################################
#
# Copyright (c) 2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
The finishing states.
"""
from __future__ import absolute_import
from __future__ import print_function

def Finish(vote_state, committed_tid_int, needs_store_commit=True):
    """
    The state we enter with tpc_finish.

    This is transient; once we successfully enter this state, we immediately return
    to the not-in-transaction state.
    """
    # Bring the load connection to current status.
    vote_state.shared_state.load_connection.rollback_quietly()
    if needs_store_commit:
        # We may have already committed the store connection, so there's
        # no point doing so again. Also no point in rolling it back either.
        txn = vote_state.shared_state.prepared_txn
        assert txn is not None
        vote_state.shared_state.adapter.txncontrol.commit_phase2(
            vote_state.shared_state.store_connection,
            txn,
            vote_state.shared_state.load_connection)

    vote_state.committing_tid_lock.release_commit_lock(
        vote_state.shared_state.store_connection.cursor)
    vote_state.shared_state.cache.after_tpc_finish(vote_state.committing_tid_lock.tid,
                                                   vote_state.shared_state.temp_storage)

    # The vote caller is responsible for releasing the shared
    # resources in vote_state.shared_state.
    # XXX: Which may not make much sense?
    return vote_state.initial_state.with_committed_tid_int(
        committed_tid_int
    )



# -*- coding: utf-8 -*-
"""
Function expressions.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from .expressions import Expression
from .schema import Column

class _Functions(object):

    def max(self, column):
        return _Function('max', column)

    def count(self, column=Column('*')):
        return _Function('COUNT', column)

class _Function(Expression):

    def __init__(self, name, expression):
        self.name = name
        self.expression = expression

    def __compile_visit__(self, compiler):
        compiler.emit_identifier(self.name)
        compiler.visit_grouped(self.expression)

func = _Functions()



##############################################################################
#
# Copyright (c) 2009,2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Blob management utilities needed by RelStorage.

Most of this code is lifted from ZODB/ZEO.
"""
from __future__ import absolute_import
from __future__ import print_function


from ZODB.POSException import Unsupported
from zope.interface import implementer

from .interfaces import INoBlobHelper


__all__ = [
    'BlobHelper',
]


@implementer(INoBlobHelper)
class NoBlobHelper(object):
    # pylint:disable=unused-argument

    __slots__ = ()

    NEEDS_DB_LOCK_TO_FINISH = False
    NEEDS_DB_LOCK_TO_VOTE = False

    shared_blob_helper = False
    txn_has_blobs = False
    shared_blob_dir = None

    def new_instance(self, adapter):
        return self

    vote = finish = lambda self, _tid=None: None
    begin = abort = clear_temp = close = lambda self: None

    def after_pack(self, oid_int, tid_int):
        """
        Because there cannot be blobs, this method has nothing to do.
        """

    def copy_undone(self, copied, tid):
        """
        Because there cannot be blobs, this method has nothing to do.
        """

    def loadBlob(self, cursor, oid, serial):
        raise Unsupported("No blob directory is configured.")

    def openCommittedBlobFile(self, cursor, oid, serial, blob=None):
        raise Unsupported("No blob directory is configured.")

    def temporaryDirectory(self):
        raise Unsupported("No blob directory is configured.")

    def storeBlob(self, cursor, store_func,
                  oid, serial, data, blobfilename, version, txn):
        raise Unsupported("No blob directory is configured.")

    def restoreBlob(self, cursor, oid, serial, blobfilename):
        raise Unsupported("No blob directory is configured.")

    @property
    def fshelper(self):
        raise AttributeError("NoBlobHelper has no 'fshelper'")

    def __repr__(self):
        return "<NoBlobHelper>"


def BlobHelper(options, adapter):
    if options is None or not options.blob_dir:
        return NoBlobHelper()

    # Prevent warnings from runpy that these were found in sys.modules
    # before executing .cached.
    from .cached import CacheBlobHelper
    from .shared import SharedBlobHelper

    if options.shared_blob_dir:
        return SharedBlobHelper(options, adapter)
    return CacheBlobHelper(options, adapter)



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

"""Integration tests for sqlite support."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import tempfile
import os.path
import unittest

from relstorage.adapters.sqlite.adapter import Sqlite3Adapter
from relstorage.options import Options

from relstorage.tests import reltestbase
from relstorage.tests.util import AbstractTestSuiteBuilder
from relstorage.tests.util import RUNNING_ON_TRAVIS
from relstorage.tests.util import RUNNING_ON_APPVEYOR
from relstorage._compat import PYPY
from relstorage._compat import PY3

class Sqlite3AdapterMixin(object):

    def __get_db_name(self):
        if self.keep_history:
            db = self.base_dbname
        else:
            db = self.base_dbname + '_hf'
        return db

    def _get_data_dir(self):
        # Our layers tend to change the temporary directory,
        # and then destroy it when the layer is torn down.
        # So our files don't persist. This can show a different set of
        # bugs than re-using an existing schema that gets zapped.
        # return "/tmp"
        return tempfile.gettempdir()

    DEFAULT_EXTRA_PRAGMAS = {}

    def __get_adapter_options(self, dbname=None):
        dbname = dbname or self.__get_db_name()
        assert isinstance(dbname, str), (dbname, type(dbname))
        data_dir = os.path.join(
            self._get_data_dir(),
            dbname
        )

        return {
            'data_dir': data_dir,
            'pragmas': self.DEFAULT_EXTRA_PRAGMAS.copy(),
        }

    def make_adapter(self, options, db=None):

        return self.get_adapter_class()(
            options=options,
            **self.__get_adapter_options(db)
        )

    def get_adapter_class(self):
        return Sqlite3Adapter

    def get_adapter_zconfig(self):
        options = self.__get_adapter_options()
        options['data-dir'] = options['data_dir']
        del options['data_dir']
        del options['pragmas']
        options['driver'] = self.driver_name
        formatted_options = '\n'.join(
            '     %s %s' % (k, v)
            for k, v in options.items()
        )

        return u"""
        <sqlite3>
            %s
            gevent_yield_interval 42
            <pragmas>
               journal_mode memory
               cache_size 8mb
            </pragmas>
        </sqlite3>
        """ % (formatted_options)

    def verify_adapter_from_zconfig(self, adapter):
        self.assertEqual(adapter.connmanager.path,
                         os.path.join(
                             self.__get_adapter_options()['data_dir'], 'main.sqlite3'))
        if 'gevent' in adapter.connmanager.driver.__name__:
            self.assertEqual(adapter.connmanager.driver.yield_to_gevent_instruction_interval,
                             42)
        conn, _ = adapter.connmanager.open()
        try:
            cur = conn.execute('pragma cache_size')
            size, = cur.fetchall()[0]
            self.assertEqual(size, 8388608)
            # But we weren't allowed to change journal_mode
            cur = conn.execute('pragma journal_mode')
            journal, = cur.fetchall()[0]
            self.assertEqual(journal, 'wal')
        finally:
            conn.close()


class TestSqlite3(Sqlite3AdapterMixin, reltestbase.RelStorageTestBase):

    DEFAULT_EXTRA_PRAGMAS = {
        # Checkpoint frequently so that we don't have to iterate too much.
        # The default checkpoint is usually 4MB.
        'wal_autocheckpoint': 2
    }

    def test_wal_does_not_grow(self):
        # issue 401 https://github.com/zodb/relstorage/issues/401
        import transaction
        from ZODB import DB

        def size_wal():
            d = self._storage._adapter.data_dir
            wal = os.path.join(d, 'main.sqlite3-wal')

            #os.system('ls -lh ' + wal)
            size = os.stat(wal).st_size
            return size


        def run_transaction(db):
            tx = transaction.begin()
            conn = db.open()
            try:
                root = conn.root()
                root['key'] = 'abcd' * 1000
                tx.commit()
            except:
                tx.abort()
                raise
            finally:
                conn.close()

        old_explicit = transaction.manager.explicit
        transaction.manager.explicit = True
        try:
            db = self._closing(DB(self._storage))
            # Open an isolated connection. If the transaction
            # manager isn't in explicit mode, it really will talk to
            # the database now, which will cause WAL to start accumulating.
            # Use explicit transaction managers to prevent that from happening
            # and keep a tight reign on transaction duration.
            isolated_txm = transaction.TransactionManager()
            isolated_txm.explicit = True # If this is commented out, the WAL grows without bound
            c1 = db.open(isolated_txm)

            try:
                TX_COUNT = 100
                RUNS = 3
                # Prime it.
                for _ in range(TX_COUNT):
                    run_transaction(db)
                wal_size = size_wal()

                for _ in range(RUNS):
                    for _ in range(TX_COUNT):
                        run_transaction(db)
                    self.assertEqual(wal_size, size_wal())
            finally:
                c1.close()
                db.close()

        finally:
            transaction.manager.explicit = old_explicit

class Sqlite3TestSuiteBuilder(AbstractTestSuiteBuilder):

    __name__ = 'Sqlite3'
    RAISED_EXCEPTIONS = Exception

    def __init__(self):
        from relstorage.adapters.sqlite import drivers
        super(Sqlite3TestSuiteBuilder, self).__init__(
            drivers,
            Sqlite3AdapterMixin,
            extra_test_classes=(TestSqlite3,)
        )

    def _compute_large_blob_size(self, use_small_blobs):
        return Options().blob_chunk_size

    __BASE_SKIPPED_TESTS = (
        # These were both seen on Travis with PyPy3.6 7.1.1, sqlite 3.11.
        # I can't reproduce locally.
        ('checkAutoReconnect', PYPY and PY3 and RUNNING_ON_TRAVIS,
         "Somehow still winds up closed"),
        ('checkAutoReconnectOnSync', PYPY and PY3 and RUNNING_ON_TRAVIS,
         "Somehow still winds up closed"),
    )

    def __add_skips(self, klass, extra_skips=()):
        for mname, skip, message in self.__BASE_SKIPPED_TESTS + extra_skips:
            meth = getattr(klass, mname)
            meth = unittest.skipIf(skip, message)(meth)
            setattr(klass, mname, meth)

    def _make_check_class_HistoryPreservingRelStorageTests(self, bases,
                                                           name, klass_dict=None):
        klass = self._default_make_check_class(bases, name, klass_dict=klass_dict)
        skips = (
            # For some reason this fails to get the undo log. Something
            # to do with the way we manage the connection? Seen on Python 3.7
            # with sqlite 3.28
            ('checkPackUnlinkedFromRoot', RUNNING_ON_APPVEYOR,
             "Fails to get undo log"),
            # Ditto, but on Python 2.7 with sqlite 3.14
            ('checkTransactionalUndoAfterPackWithObjectUnlinkFromRoot', RUNNING_ON_APPVEYOR,
             "Fails to get undo log"),
        )
        self.__add_skips(klass, skips)
        return klass

    def _make_check_class_HistoryFreeRelStorageTests(self, bases,
                                                     name, klass_dict=None):
        klass = self._default_make_check_class(bases, name, klass_dict=klass_dict)
        self.__add_skips(klass)
        return klass

def test_suite():
    return Sqlite3TestSuiteBuilder().test_suite()



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
The definition of how local cache databases are manipulated.

This is the bare bones byte-shuffling layer, it defines as little policy
as possible (while keeping in mind its purpose.)

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from abc import abstractmethod
from contextlib import closing
import sqlite3

from relstorage._compat import ABC
from relstorage._compat import OID_TID_MAP_TYPE
from relstorage._util import log_timed

from relstorage.adapters.sqlite.batch import Sqlite3RowBatcher
from relstorage.adapters.sqlite.dialect import SQ3_SUPPORTS_UPSERT as SUPPORTS_UPSERT


logger = __import__('logging').getLogger(__name__)

class SimpleQueryProperty(object):
    """
    Wraps a query that returns one value in one row.
    """

    def __init__(self, sql, param_names=()):
        self.query = sql
        self.param_names = param_names

    def __get__(self, inst, klass):
        if inst is None: # pragma: no cover
            return self

        cur = inst.connection.execute(
            self.query,
            [getattr(inst, n) for n in self.param_names])

        result = cur.fetchone()[0]
        cur.close()
        return result



class Database(ABC):
    """
    The database stores only one state for each object.

    This should generally be the latest state found in the cache.
    """

    @classmethod
    def from_connection(cls,
                        connection,
                        use_upsert=SUPPORTS_UPSERT):

        kind = _UpsertUpdateDatabase if use_upsert else _InsertReplaceDatabase
        return kind(connection)

    def __init__(self, connection):
        self.connection = connection
        self.cursor = connection.cursor()
        self.cursor.arraysize = 100
        self.create_schema()

    def close(self):
        try:
            if self.cursor is not None:
                self.cursor.close()
            if self.connection is not None:
                self.connection.close()
        finally:
            self.cursor = None
            self.connection = None

    # The main repository of our data. This uses the OID of the object
    # as the INTEGER PRIMARY KEY --- that's a special type of key that
    # means this table is a clustered table, organized with that
    # column as its primary key.
    #
    # This reduces overhead of having a secondary (hidden) 'rowid' column
    # to do the clustering on, to it's important.
    _state_table_schema = """
    CREATE TABLE IF NOT EXISTS object_state (
        zoid INTEGER PRIMARY KEY,
        tid INTEGER NOT NULL ,
        was_frozen INT(1) NOT NULL DEFAULT 0,
        frequency INTEGER NOT NULL,
        state BLOB
    );
    """

    # We want to keep the clustering for the temporary table,
    # so the integer primary key matters.
    _temp_table_schema = _state_table_schema.replace(
        "object_state", 'temp_state'
    ).replace('TABLE', 'TEMPORARY TABLE')

    _schema = _state_table_schema + '\n' + _temp_table_schema + """
    CREATE TABLE IF NOT EXISTS checkpoints (
        id INTEGER PRIMARY KEY, max_hvt INTEGER, complete_since INTEGER
    );

    CREATE INDEX IF NOT EXISTS IX_object_state_f_tid
    ON object_state (frequency DESC, tid DESC);
    """

    # Without the CAST AS BLOB, if a value went in with text affinity,
    # (which happens essentially always under Python 2 but if we've done
    # things right never under Python 3) LENGTH will stop at an embedded
    # NUL. Those are common in pickles, so our size calculation will be
    # very wrong.
    total_state_len = SimpleQueryProperty(
        "SELECT TOTAL(LENGTH(CAST(state AS blob))) FROM object_state"
    )

    total_state_count = SimpleQueryProperty(
        "SELECT COUNT(zoid) FROM object_state"
    )

    def create_schema(self):
        self.cursor.executescript(self._schema)

    @log_timed
    def _read_oids_and_tids_from_db_to_map(self):
        cur = self.connection.execute('SELECT zoid, tid FROM object_state')
        with closing(cur):
            return OID_TID_MAP_TYPE(cur.fetchall())

    oid_to_tid = property(_read_oids_and_tids_from_db_to_map,
                          doc="""
        A map from OID to its corresponding TID, for
        all the data in the database.
        """)

    @property
    def checkpoints(self):
        """
        The checkpoints in the database, or None if there are none.
        """
        self.cursor.execute("SELECT max_hvt, complete_since FROM checkpoints")
        return self.cursor.fetchone()

    def _remove_invalid_persistent_oids(self, bad_oids, cur):
        cur.execute("BEGIN")
        batch = Sqlite3RowBatcher(cur)
        for oid in bad_oids:
            batch.delete_from('object_state', zoid=oid)
        batch.flush()
        cur.execute("COMMIT")
        return batch.total_rows_deleted

    def remove_invalid_persistent_oids(self, bad_oids):
        # The database might be locked by others, either someone in
        # this method or someone actually closing the cache and
        # writing to the database, a situation we won't detect until
        # we try to actually remove a row (it's entirely possible the
        # rows we want to remove are already gone, so we don't BEGIN
        # IMMEDIATE to force the issue.)
        #
        # Our workaround is to try a few times and then give up. Longer timeouts
        # make it less likely we need to do this.
        tries = 3
        cur = self.cursor
        while tries:
            tries -= 1
            try:
                return self._remove_invalid_persistent_oids(bad_oids, cur)
            except sqlite3.OperationalError:
                # If we don't rollback, we get 'cannot BEGIN inside a transaction'
                cur.execute('ROLLBACK')
                # No need to sleep, that's built in to the timeout parameter
                # when we connect.
                logger.debug("Failed to lock database to remove OIDs; tries left: %d", tries)
        return -1

    def fetch_rows_by_priority(self):
        """
        The returned object will iterate ``(zoid, was_frozen, state, tid, frequency)``
        from most frequently used and newest, to least frequently used and oldest.

        You *must* completely consume the returned object.
        """
        # Do this in a new cursor so it can interleave.

        # Read these in priority order; as a tie-breaker, choose newer transactions
        # over older transactions.
        # We could  use a window function over SUM(LENGTH(state)) to only select
        # the rows that will actually fit:
        #
        # SELECT * FROM (
        #  SELECT zoid, tid, state,
        #   sum(length(state)) over (order by frequency desc, tid desc) as cum_size
        #  FROM object_state
        # ORDER BY frequency DESC, tid DESC
        # )
        # WHERE cum_size < ?
        #
        # However, that seems to generate a poor query plan that actually looks
        # at all the rows (it doesn't understand that cum_size can only increase.)
        # Plus, window functions were only added to sqlite 3.25
        as_state = bytes # Py2 returns buffers, Py3 returns bytes. bytes(bytes) is a no-op.
        cur = self.connection.cursor()
        cur.arraysize = 100
        cur.execute("""
            SELECT zoid, CASE was_frozen WHEN 1 THEN -1 ELSE tid END,
                   CAST(state AS BLOB), tid, frequency
            FROM object_state
            ORDER BY frequency DESC, tid DESC
        """)
        for zoid, frozen, state, tid, frequency in cur:
            yield zoid, frozen, as_state(state), tid, frequency

    @log_timed
    def list_rows_by_priority(self):
        """
        Like ``fetch_rows_by_priority``, but returns a list instead of cursor.
        """
        return list(self.fetch_rows_by_priority())

    def store_temp(self, rows):
        """
        Given an iterator of ``(oid, tid, frozen, state, frequency)`` values,
        store them in a temporary table for this session.
        """
        # The batch size depends on how many params a stored proc can
        # have; if we go too big we get OperationalError: too many SQL
        # variables. The default is 999.
        # Note that the multiple-value syntax was added in
        # 3.7.11, 2012-03-20.

        # Benchmarking shows essentially no difference between this
        # simple method and using our RowBatcher to produce
        # multi-value statements. vmprof shows all of the time spent
        # in *this* function right here, nothing any lower (the next
        # lower function it shows is _pysqlite_fetch_one_row, taking
        # 1.1% of the execution of *this* function). I'm Not entirely
        # sure what that means.

        # Because of sqlite's loose typing. on Python 2 we may get
        # state column values stored with TEXT affinity, even though
        # the storage class of the state column is BLOB. On Python 3
        # reading that back will try to decode as UTF-8 and result in
        # decode errors:
        #
        # sqlite3.OperationalError: Could not decode to UTF-8 column
        # 'state' with text '.zx....'
        #
        # We can add a CAST(state as BLOB), or we could set the
        # connection's text_factory to bytes (which makes the metadata
        # bytes too).
        rows = list(rows) # materialize
        self.cursor.executemany(
            'INSERT INTO temp_state(zoid, tid, was_frozen, state, frequency) '
            'VALUES (?, ?, ?, ?, ?)',
            rows
        )

        return len(rows), -1

    @abstractmethod
    def move_from_temp(self):
        """
        Take rows in the temporary table and put them in the permanent table,
        overwriting rows for the same object that are older (based on TID)

        If there is a row that is newer, then it is preserved and the temporary
        row is discarded.

        The temporary table will be clear after this.

        Returns the total number of rows that were stored into the permanent table.
        """
        raise NotImplementedError

    @abstractmethod
    def update_checkpoints(self, cp0, cp1):
        """
        Save these checkpoints, if they are newer than the current checkpoints.
        """
        raise NotImplementedError

    def trim_to_size(self, limit, min_allowed_oid):
        # Manipulates a transaction.
        if not min_allowed_oid and self.total_state_len <= limit:
            # Nothing to do.
            return

        # Take out the lock and check again.
        cur = self.cursor
        if min_allowed_oid:
            # This could be easily optimized for a small number of rows,
            # or use a custom RowBatcher that handles <= instead of =
            # operator.
            logger.info("Checking table of size %d against %d stale entries",
                        self.total_state_count, len(min_allowed_oid))
            def is_stale(zoid, tid, min_allowed=min_allowed_oid.get):
                return min_allowed(zoid, tid) > tid

            self.connection.create_function('is_stale', 2, is_stale)

        cur.execute('BEGIN IMMEDIATE')

        if min_allowed_oid:
            cur.execute('DELETE FROM object_state WHERE is_stale(zoid, tid)')

        byte_count = self.total_state_len
        if byte_count <= limit:
            # Someone else did it, yay!
            cur.execute('COMMIT')
            return

        really_big = byte_count > limit * 2
        how_much_to_trim = byte_count - limit
        logger.info(
            "State too large; need to trim %d to reach %d",
            how_much_to_trim,
            limit,
        )

        rows_deleted = self._trim_state(how_much_to_trim)

        cur.execute('COMMIT')
        # Rewrite the file? If we were way over our target, that probably
        # matters. And sometimes we might want to do it just to do it and
        # optimize the tables.
        if really_big:
            cur.execute('VACUUM')
        logger.info(
            "Trimmed %d rows (desired: %d actual: %d)",
            rows_deleted, limit, self.total_state_len
        )


    def _trim_state(self, how_much_to_trim):
        # Try to get the oldest, least used, biggest objects we can.
        # We could probably use a window function over SUM(LENGTH(state))
        # to limit the select to just the rows we want.

        # We'll be interleaving statements so we must use a
        # separate cursor
        batch_cur = self.connection.cursor()
        batch = Sqlite3RowBatcher(batch_cur)

        # In fact, because of the way PyPy wants you to fetch all rows
        # or it considers some statements to still be open and thus
        # refuses to allow things like VACUUM, we need to use two
        # cursors, so we can close the fetch cursor too
        fetch_cur = self.connection.cursor()
        fetch_cur.execute("""
        SELECT zoid, LENGTH(state)
        FROM object_state
        ORDER BY frequency ASC, tid ASC, LENGTH(CAST(state AS BLOB)) DESC, zoid ASC
        """)


        for row in fetch_cur:
            zoid, size = row
            how_much_to_trim -= size
            batch.delete_from('object_state', zoid=zoid)
            if how_much_to_trim <= 0:
                break
        batch.flush()
        batch_cur.close()
        fetch_cur.close()
        return batch.total_rows_deleted

class _UpsertUpdateDatabase(Database):

    def move_from_temp(self):
        # "The parser might not be able to tell if the "ON" keyword is
        # introducing the UPSERT or if it is the ON clause of a join.
        # To work around this, the SELECT statement should always
        # include a WHERE clause, even if that WHERE clause is just
        # ``WHERE true``."
        self.cursor.execute("""
        INSERT INTO object_state (zoid, tid, was_frozen, frequency, state)
        SELECT zoid, tid, was_frozen, frequency, state
        FROM temp_state
        WHERE true
        ON CONFLICT(zoid) DO UPDATE
        SET tid = excluded.tid,
            was_frozen = excluded.was_frozen,
            state = excluded.state,
            frequency = excluded.frequency + object_state.frequency
        WHERE excluded.tid >= tid
        """)
        rows_inserted = self.cursor.rowcount
        self.cursor.execute("DELETE FROM temp_state")
        return rows_inserted

    def update_checkpoints(self, cp0, cp1):
        self.cursor.execute("""
        INSERT INTO checkpoints (id, max_hvt, complete_since)
        VALUES (0, ?, ?)
        ON CONFLICT(id) DO UPDATE
            SET max_hvt = excluded.max_hvt, complete_since = excluded.complete_since
        WHERE excluded.max_hvt >= max_hvt
        """, (cp0, cp1))

class _InsertReplaceDatabase(Database):
    def move_from_temp(self):
        # The old 'INSERT OR REPLACE' syntax is supported from
        # 3.0.0 forward. It's not as flexible as the true upsert:
        # for example, you can't specify the type of conflict, nor
        # can you use a WHERE clause or specify the values to use
        # in the update. It also doesn't increment the cursor's change
        # counter for replaced rows.
        self.cursor.execute("""
        DELETE FROM temp_state
        WHERE EXISTS (
            SELECT 1
            FROM object_state
            WHERE object_state.zoid = temp_state.zoid
            AND object_state.tid > temp_state.tid
        )
        """)
        self.cursor.execute('SELECT COUNT(*) FROM temp_state')
        rows_inserted = self.cursor.fetchall()[0][0]

        self.cursor.execute("""
        INSERT OR REPLACE INTO object_state (zoid, tid, was_frozen, state, frequency)
        SELECT zoid, tid, was_frozen, state, frequency
        FROM temp_state
        """)

        self.cursor.execute("DELETE FROM temp_state")
        return rows_inserted

    def update_checkpoints(self, cp0, cp1):
        cur = self.cursor
        cur.execute("""
        INSERT OR REPLACE INTO checkpoints(id, max_hvt, complete_since)
        SELECT 0, ?, ?
        FROM checkpoints
        WHERE max_hvt <= ?
        UNION
        SELECT 0, ?, ?
        WHERE NOT EXISTS (SELECT id FROM checkpoints)
        """, (
            cp0, cp1,
            cp0,
            cp0, cp1
        ))



# -*- coding: utf-8 -*-
"""
Concepts that exist in SQL, such as columns and tables.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from relstorage._util import Lazy

from ._util import Columns
from ._util import copy

from .types import Type
from .types import Unknown

from .expressions import ExpressionOperatorMixin
from .expressions import ParamMixin

from .dialect import DialectAware

from .select import Selectable
from .insert import Insertable
from .insert import Deletable
from .insert import Updatable
from .insert import Upsertable
from .query import Query

_marker = object()

class Column(ExpressionOperatorMixin):
    """
    Defines a column in a table.
    """
    # If present, the table name of the table we were defined in.
    table = None

    def __init__(self, name, type_=Unknown,
                 primary_key=False,
                 nullable=True,
                 auto_increment=False,
                 default=_marker):
        self.name = name
        self.type_ = type_ if isinstance(type_, Type) else type_()
        self.primary_key = primary_key
        self.nullable = False if primary_key else nullable
        self.auto_increment = auto_increment
        self.default = default
        self.alias = name

    def is_type(self, type_):
        return isinstance(self.type_, type_)

    def resolve_against(self, table):
        if table.c.is_ambiguous(self):
            return _QualifiedColumn(self)
        return self

    def aliased(self, alias):
        return _AliasedColumn(self, alias)

    def __str__(self):
        return self.name

    def __repr__(self):
        result = '%s(%r,' % (
            type(self).__name__,
            self.name,
        )
        if self.table:
            result += 'table=%r' % (self.table,)
        result += ')'
        return result

    def __compile_visit__(self, compiler):
        compiler.visit_column(self)

    def __compile_visit_for_create__(self, compiler):
        compiler.visit_column(self)
        compiler.emit_keyword(
            compiler.dialect.datatype_for_column(self)
        )
        if self.primary_key:
            compiler.emit_column_constraint('PRIMARY KEY')
        # Order matters for some databases.
        # - In Oracle, the DEFAULT has to come before NOT NULL
        # - In sqlite, the auto_increment has to be before NOT NULL (?)
        if self.default is not _marker:
            compiler.emit_keyword('DEFAULT')
            compiler.visit_immediate_expression(self.default)

        if self.auto_increment:
            compiler.emit_column_autoincrement()
        if not self.nullable:
            compiler.emit_column_constraint('NOT NULL')
        compiler.emit(
            compiler.dialect.extra_constraints_for_column(self)
        )

class ColumnExpression(Column):

    def resolve_against(self, table):
        return self

    def __str__(self):
        return str(self.name)

    def aliased(self, alias):
        return _AliasedColumnExpression(self, alias)

class _AliasedColumn(Column):
    def __init__(self, other, alias):
        Column.__init__(self, other.name)
        vars(self).update(vars(other))
        self.alias = alias

    def __compile_visit__(self, compiler):
        compiler.visit_aliased_column(self)

class _AliasedColumnExpression(_AliasedColumn):

    def __str__(self):
        return str(self.name)

class _QualifiedColumn(DialectAware,
                       Column):

    def __init__(self, other):
        Column.__init__(self, other.name)
        vars(self).update(vars(other))
        assert self.table

    def __str__(self):
        return "%s.%s" % (self.table, self.name)

    def __compile_visit__(self, compiler):
        compiler.visit_qualified_column(self)

class _DeferredColumn(Column):

    def resolve_against(self, table):
        col = getattr(table.c, self.name)
        if col is not self:
            # We may not be able to resolve fully until the table
            # is bound.
            col = col.resolve_against(table)
        return col

class _DeferredColumns(object):

    def __getattr__(self, name):
        return _DeferredColumn(name)


class ColumnResolvingProxy(ParamMixin):
    """
    A proxy that select can resolve to tables in the current table.
    """

    c = _DeferredColumns()


class SchemaItem(object):
    """
    A permanent item in a schema (aka the data dictionary).
    """


class _CreateTable(Query):

    def __init__(self, table, **kw):
        self.table = table
        self.kw = kw

    def __compile_visit__(self, compiler):
        compiler.create_table(self.table, **self.kw)


class Table(Selectable,
            Insertable,
            Deletable,
            Updatable,
            Upsertable,
            ParamMixin,
            SchemaItem):
    """
    A table relation.
    """

    def __init__(self, name, *columns):
        self.name = name
        self.columns = columns
        for c in columns:
            c.table = name
        self.c = Columns(columns)

    def __str__(self):
        return self.name

    def __repr__(self):
        return '%s(%r,%s)' % (
            type(self).__name__,
            self.name,
            ','.join('Column(%r)' % (c.name) for c in self.columns)
        )

    def __compile_visit__(self, compiler):
        compiler.emit_identifier(self.name)
        return self

    def _emit_create_prefix(self, compiler):
        compiler.emit_keyword('CREATE')
        compiler.emit_keyword('TABLE')

    def _emit_identifier_suffix(self, compiler):
        pass

    def __compile_visit_for_create__(self, compiler, if_not_exists=False):
        self._emit_create_prefix(compiler)
        if if_not_exists:
            compiler.emit_if_not_exists()
        compiler.emit_identifier(self.name)
        self._emit_identifier_suffix(compiler)
        self._visit_definition(compiler)
        return self

    def _visit_definition(self, compiler):
        compiler.visit_grouped(
            self.c
        )

    def natural_join(self, other_table):
        return NaturalJoinedTable(self, other_table)

    def inner_join(self, other_table):
        return _InnerJoinBuilder(self, other_table)

    def resolve_against(self, _table):
        return self

    def create(self, if_not_exists=False):
        """
        WARNING: This currently doesn't support table constraints
        or foreign keys.
        """
        return _CreateTable(self, if_not_exists=if_not_exists)


class TemporaryTable(Table):
    """
    A temporary table.
    """
    def _emit_create_prefix(self, compiler):
        compiler.emit_keyword('CREATE')
        compiler.emit_keyword('TEMPORARY')
        compiler.emit_keyword('TABLE')

class View(DialectAware,
           Table):
    """
    A table defined by a query.
    """

    def __init__(self, name, select):
        # The column names can't be qualified against the original table,
        # they disappear, so we need to rebind them against a new table.
        columns = [copy(c) for c in select.column_list]
        super(View, self).__init__(name, *columns)
        self.select = select

    def _emit_create_prefix(self, compiler):
        compiler.emit_keyword('CREATE')
        compiler.emit_keyword('VIEW')

    def _emit_identifier_suffix(self, compiler):
        compiler.emit_keyword('AS')

    def _visit_definition(self, compiler):
        with compiler.using_visit_name(''):
            compiler.visit(self.select)

    def __str__(self):
        # This is almost how you would inline it in a FROM clause
        return '(%s)' % (self.select,)

    def create(self): # pylint:disable=arguments-differ
        return _CreateTable(self, if_not_exists=False, use_trailer=False)


class _CompositeTableMixin(DialectAware,
                           Selectable,
                           ParamMixin):
    """
    A virtual table whose columns are the common set of columns
    found in two tables.
    """
    def __init__(self, lhs, rhs):
        self.lhs = lhs
        self.rhs = rhs

    _bind_vars_ignored = ('c', 'columns')

    def _bound_to(self, context, dialect):
        for v in self._bind_vars_ignored:
            self.__dict__.pop(v, None)
        return super(_CompositeTableMixin, self)._bound_to(context, dialect)

    @Lazy
    def c(self):
        return self.lhs.c.intersection(self.rhs.c)

    @Lazy
    def columns(self):
        return tuple(self.c)

    def resolve_against(self, _):
        return self

    def __repr__(self):
        return '<%s at 0x%x lhs=%r rhs=%r>' % (
            type(self).__name__,
            id(self),
            self.lhs,
            self.rhs,
        )

class _JoinedTable(_CompositeTableMixin):

    @Lazy
    def c(self):
        # Make the union of columns available,
        # being sure that if we're doing a USING join
        # we don't double the list of columns we're USING,
        # because we shouldn't use prefixes for those
        # (Oracle doesn't like it)
        join_columns = self._get_join_columns(
            self.lhs.resolve_against(self),
            self.rhs.resolve_against(self))
        union = self.lhs.c.dup_union(self.rhs.c, combining=join_columns.as_names())

        return union

    _JOIN_KW = 'JOIN'
    _JOIN_KIND_KW = 'USING'

    def _get_join_columns(self, lhs, rhs):
        return lhs.c.intersection(rhs.c)

    def __compile_visit__(self, compiler):
        # In case we're wrapped around HistoryVariantTable,
        # defer selection of join columns until now.
        # TODO: This probably only handles one extra level of joins?
        lhs = self.lhs.resolve_against(self)
        rhs = self.rhs.resolve_against(self)
        if isinstance(rhs, _JoinedTable):
            rhs = compiler.visit(rhs)
            compiler.emit_keyword(self._JOIN_KW)
            lhs = compiler.visit(lhs)
            compiler.emit_keyword(self._JOIN_KIND_KW)
            compiler.visit_grouped(self._get_join_columns(lhs, rhs))
        else:
            lhs = compiler.visit(self.lhs)
            compiler.emit_keyword(self._JOIN_KW)
            rhs = compiler.visit(self.rhs)
            compiler.emit_keyword(self._JOIN_KIND_KW)
            compiler.visit_grouped(self._get_join_columns(lhs, rhs))
        return self

    def join_kind(self, kind):
        s = copy(self)
        s._JOIN_KW = kind
        return s

    def __repr__(self):
        return '<%s at 0x%x %r %s %r (%r)>' % (
            type(self).__name__,
            id(self),
            self.lhs,
            self._JOIN_KW,
            self.rhs,
            self.c
        )

class NaturalJoinedTable(_JoinedTable):
    pass

class _InnerJoinBuilder(object):

    def __init__(self, lhs, rhs):
        self.lhs = lhs
        self.rhs = rhs

    def using(self, *columns):
        return UsingJoinedTable(self.lhs, self.rhs, columns)

class UsingJoinedTable(_JoinedTable):

    def __init__(self, lhs, rhs, columns):
        super(UsingJoinedTable, self).__init__(lhs, rhs)
        self._join_columns = Columns(columns)

    def _get_join_columns(self, lhs, rhs):
        c = self._join_columns
        assert self.lhs.c.intersection(c) == c
        assert self.rhs.c.intersection(c) == c
        return c


class HistoryVariantTable(_CompositeTableMixin):
    """
    A table that can be one of two tables, depending on whether
    the instance is keeping history or not.
    """

    @Lazy
    def c(self):
        # Lose the owning table of the columns so we can write the correct
        # one when bound
        union = _CompositeTableMixin.c.data[0](self) # pylint:disable=no-member
        return Columns([
            _DeferredColumn(c.name) for c in union])


    def _bound_to(self, context, dialect):
        t = self.history_preserving if context.keep_history else self.history_free
        return t.bind(context, dialect) if hasattr(t, 'bind') else t

    @property
    def history_preserving(self):
        return self.lhs

    @property
    def history_free(self):
        return self.rhs

    @property
    def table_in_context(self):
        keep_history = self.context.keep_history
        node = self.history_preserving if keep_history else self.history_free
        return node

    def resolve_against(self, _):
        return self.table_in_context

    def __compile_visit__(self, compiler):
        return compiler.visit(self.table_in_context)
